﻿using System;
using System.Collections.Generic;
using System.Collections.Specialized;
using System.Data;
using System.Data.OleDb;
using System.IO;
using System.Text;
using System.Xml;
using MySql.Data.MySqlClient;
using OAI.Server.Catalog;
using OAI.Server.Verb;
using OAI;
using OAI.Util;
using ServiceCore.ServiceComponents;
using Connection = ServiceCore.ServiceComponents.Connection;

namespace ServiceCore {
	public class ServiceEngine : IOAIHandler {
		public static ServiceEngine Instance { get { return _instance; } }

		private static readonly ServiceEngine _instance = new ServiceEngine();

		public NameValueCollection Query { get; set; }
		public string Institution { get; set; }
		public string Service { get; set; }
		public static string PropertiesPath { get; set; }
		public string BaseURL { get; set; }
		public string ContentType { get; set; }

		public string Granularity { get; private set; }
		public List<string> AdminEmails { get; private set; }

		private static readonly Dictionary<string, Connection> _connections = new Dictionary<string, Connection>();
		private static readonly Dictionary<string, Service> _services = new Dictionary<string, Service>();
		private static readonly Dictionary<string, string> _templates = new Dictionary<string, string>();
		private static readonly Dictionary<string, string> _sql = new Dictionary<string, string>();
		private static readonly Dictionary<string, IDbConnection> _connectionCache = new Dictionary<string, IDbConnection>();

		public void LoadConfiguration(bool clear) {
			if (!clear && _services.Count > 0) return;

			_connections.Clear();
			_services.Clear();
			_templates.Clear();
			_connectionCache.Clear();
			_sql.Clear();

			var xd = new XmlDocument();

			try {
				xd.Load(string.Format("{0}\\{1}", PropertiesPath, SettingsConfigFile));
			} catch (XmlException) {
				throw new ServiceException(string.Format("{0} does not have well-formed xml", SettingsConfigFile));
			} catch (Exception ex) {
				throw new ServiceException(string.Format("could not load {0}: {1}", SettingsConfigFile, ex));
			}

			var settings = xd.GetElementsByTagName("settings");

			if (settings == null || settings.Count == 0)
				throw new ServiceException(string.Format("no 'settings' element found in {0}", SettingsConfigFile));

			if (settings.Count > 1)
				throw new ServiceException(string.Format("too many 'settings' elements found in {0}", SettingsConfigFile));

			var xeSettings = (XmlElement)settings[0];

			var granularity = xeSettings.GetElementsByTagName("granularity");

			if (granularity == null || granularity.Count == 0)
				throw new ServiceException(string.Format("no 'granularity' element found in {0}/settings", SettingsConfigFile));

			Granularity = granularity[0].InnerText;

			var adminEmails = xeSettings.GetElementsByTagName("adminEmail");

			if (adminEmails == null || adminEmails.Count == 0)
				throw new ServiceException(string.Format("no 'adminEmail' elements found in {0}/settings", SettingsConfigFile));

			AdminEmails = new List<string>();

			foreach (XmlElement xeAdminEmail in adminEmails) {
				AdminEmails.Add(xeAdminEmail.InnerText);
			}

			var serviceConfigurations = xeSettings.GetElementsByTagName("serviceConfiguration");

			if (serviceConfigurations == null || serviceConfigurations.Count == 0)
				throw new ServiceException(string.Format("no 'serviceConfiguration' elements found in {0}/settings", SettingsConfigFile));

			foreach (XmlElement xeServiceConfiguration in serviceConfigurations) {
				LoadServices(xeServiceConfiguration.InnerText);
			}
		}

		private static void LoadServices(string configuration) {
			var xd = new XmlDocument();

			try {
				xd.Load(string.Format("{0}\\{1}", PropertiesPath, configuration));
			} catch (XmlException) {
				throw new ServiceException(string.Format("{0} does not have well-formed xml", configuration));
			} catch (Exception) {
				throw new ServiceException(string.Format("could not load {0}", configuration));
			}

			var serviceConfigurations = xd.GetElementsByTagName("service_configuration");

			if (serviceConfigurations == null || serviceConfigurations.Count == 0)
				throw new ServiceException(string.Format("no 'service_configuration' element found in {0}", configuration));

			if (serviceConfigurations.Count > 1)
				throw new ServiceException(string.Format("too many 'service_configuration' elements found in {0}", configuration));

			var xeServiceConfiguration = (XmlElement)serviceConfigurations[0];

			var connections = xeServiceConfiguration.GetElementsByTagName("connections");

			if (connections == null || connections.Count == 0)
				throw new ServiceException(string.Format("no 'connections' element found in {0}/service_configuration", configuration));

			if (connections.Count > 1)
				throw new ServiceException(string.Format("too many 'connections' elements found in {0}/service_configuration", configuration));

			var xeConnections = (XmlElement)connections[0];

			if (!xeConnections.HasChildNodes)
				throw new ServiceException(string.Format("no 'connection' elements found in {0}/service_configuration/connections", configuration));

			foreach (XmlElement xeConnection in xeConnections.ChildNodes) {
				if (!xeConnection.HasAttribute("id"))
					throw new ServiceException(string.Format("no 'id' attribute found in {0}/service_configuration/connections/connection[?]", configuration));

				var id = xeConnection.GetAttribute("id");

				if (_connections.ContainsKey(id))
					throw new ServiceException(string.Format("connection {0} defined twice", id));

				if (!xeConnection.HasAttribute("connectionString"))
					throw new ServiceException(string.Format("no 'connectionString' attribute found in {0}/service_configuration/connections/connection[{1}]", configuration, id));

				if (!xeConnection.HasAttribute("type"))
					throw new ServiceException(string.Format("no 'type' attribute found in {0}/service_configuration/connections/connection[{1}]", configuration, id));

				var connection = new Connection(
					id,
					xeConnection.GetAttribute("connectionString"),
					xeConnection.GetAttribute("type"));
				_connections[connection.Id] = connection;
			}

			var services = xeServiceConfiguration.GetElementsByTagName("services");

			if (services == null || services.Count == 0)
				throw new ServiceException(string.Format("no 'services' element found in {0}/service_configuration", configuration));

			if (services.Count > 1)
				throw new ServiceException(string.Format("too many 'services' elements found in {0}/service_configuration", configuration));

			var xeServices = (XmlElement)services[0];

			if (!xeServices.HasChildNodes)
				throw new ServiceException(string.Format("no 'service' elements found in {0}/service_configuration/services", configuration));

			foreach (XmlElement xeService in xeServices.ChildNodes) {
				if (!xeService.HasAttribute("id"))
					throw new ServiceException(string.Format("no 'id' attribute found in {0}/service_configuration/services/service[?]", configuration));

				var id = xeService.GetAttribute("id");

				if (!xeService.HasAttribute("name"))
					throw new ServiceException(string.Format("no 'name' attribute found in {0}/service_configuration/services/service[{1}]", configuration, id));

				var name = xeService.GetAttribute("name");

				if (!xeService.HasAttribute("institution"))
					throw new ServiceException(string.Format("no 'institution' attribute found in {0}/service_configuration/services/service[{1}]", configuration, id));

				var institution = xeService.GetAttribute("institution");

				if (_services.ContainsKey(string.Format("{0}/{1}", institution.ToLower(), id.ToLower())))
					throw new ServiceException(string.Format("service {0}/{1} defined twice", institution.ToLower(), id.ToLower()));

				if (!xeService.HasAttribute("connection"))
					throw new ServiceException(string.Format("no 'connection' attribute found in {0}/service_configuration/services/service[{1}]", configuration, id));

				var connection = _connections[xeService.GetAttribute("connection")];

				if (!xeService.HasAttribute("update_rate"))
					throw new ServiceException(string.Format("no 'update_rate' attribute found in {0}/service_configuration/services/service[{1}]", configuration, id));

				var updateRate = int.Parse(xeService.GetAttribute("update_rate"));

				if (!xeService.HasAttribute("result_block"))
					throw new ServiceException(string.Format("no 'result_block' attribute found in {0}/service_configuration/services/service[{1}]", configuration, id));

				var resultBlock = int.Parse(xeService.GetAttribute("result_block"));

				if (!xeService.HasAttribute("deleted_record"))
					throw new ServiceException(string.Format("no 'deleted_record' attribute found in {0}/service_configuration/services/service[{1}]", configuration, id));

				var deletedRecord = xeService.GetAttribute("deleted_record");

				if (!xeService.HasAttribute("earliest_datestamp"))
					throw new ServiceException(string.Format("no 'earliest_datestamp' attribute found in {0}/service_configuration/services/service[{1}]", configuration, id));

				var earliestDatestamp = xeService.GetAttribute("earliest_datestamp");


				if (!xeService.HasAttribute("identify_prefix"))
					throw new ServiceException(string.Format("no 'identify_prefix' attribute found in {0}/service_configuration/services/service[{1}]", configuration, id));

				var identifyPrefix = xeService.GetAttribute("identify_prefix");

				if (!xeService.HasChildNodes)
					throw new ServiceException(string.Format("no 'protocol' elements found in {0}/service_configuration/services/service[{1}]", configuration, id));

				var commandTimeout = 0;

				if (xeService.HasAttribute("command_timeout"))
					commandTimeout = int.Parse(xeService.GetAttribute("command_timeout"));

				if (!xeService.HasAttribute("identifier_type"))
					throw new ServiceException(string.Format("no 'identifier_type' attribute found in {0}/service_configuration/services/service[{1}]", configuration, id));

				var identifierType = xeService.GetAttribute("identifier_type");

				var service = new Service(id, name, institution, connection, updateRate, resultBlock, deletedRecord, earliestDatestamp, identifyPrefix, commandTimeout, identifierType);
				_services[string.Format("{0}/{1}", institution.ToLower(), id.ToLower())] = service;

				foreach (XmlElement xeProtocol in xeService.ChildNodes) {
					if (!xeProtocol.HasAttribute("id"))
						throw new ServiceException(string.Format("no 'id' attribute found in {0}/service_configuration/services/service[{1}]/protocols/protocol[?]", configuration, id));

					var protocolId = xeProtocol.GetAttribute("id");

					if (!xeProtocol.HasAttribute("configuration"))
						throw new ServiceException(string.Format("no 'configuration' attribute found in {0}/service_configuration/services/service[{1}]/protocols/protocol[{2}]", configuration, id, protocolId));

					var protocolConfig = xeProtocol.GetAttribute("configuration");

					if (!xeProtocol.HasAttribute("schema"))
						throw new ServiceException(string.Format("no 'schema' attribute found in {0}/service_configuration/services/service[{1}]/protocols/protocol[{2}]", configuration, id, protocolId));

					var schema = xeProtocol.GetAttribute("schema");

					if (!xeProtocol.HasAttribute("namespace"))
						throw new ServiceException(string.Format("no 'namespace' attribute found in {0}/service_configuration/services/service[{1}]/protocols/protocol[{2}]", configuration, id, protocolId));

					var ns = xeProtocol.GetAttribute("namespace");

					if (!xeProtocol.HasAttribute("content-type"))
						throw new ServiceException(string.Format("no 'content-type' attribute found in {0}/service_configuration/services/service[{1}]/protocols/protocol[{2}]", configuration, id, protocolId));

					var contentType = xeProtocol.GetAttribute("content-type");

					var constants = new Dictionary<string, Constant>();

					if (xeProtocol.HasChildNodes) {
						foreach (XmlElement xeConstant in xeProtocol.ChildNodes) {
							if (!xeConstant.HasAttribute("id"))
								throw new ServiceException(string.Format("no 'id' attribute found in {0}/service_configuration/services/service[{1}]/protocols/protocol[{2}]/constant[?]", configuration, id, protocolId));

							var constantId = xeConstant.GetAttribute("id");

							if (!xeConstant.HasAttribute("value"))
								throw new ServiceException(string.Format("no 'value' attribute found in {0}/service_configuration/services/service[{1}]/protocols/protocol[{2}]/constant[{3}]", configuration, id, protocolId, constantId));

							constants[constantId] = new Constant(constantId, xeConstant.GetAttribute("value"));
						}
					}

					service.Descriptions[protocolId] = GetProtocolDescription(protocolConfig, schema, ns, contentType, constants);
				}

			}
		}

		private static ProtocolDescription GetProtocolDescription(string protocolPath, string schema, string ns, string contentType, Dictionary<string, Constant> constants) {
			var xd = new XmlDocument();

			try {
				xd.Load(string.Format("{0}\\{1}", PropertiesPath, protocolPath));
			} catch (XmlException) {
				throw new ServiceException(string.Format("{0} does not have well-formed xml", protocolPath));
			} catch (Exception) {
				throw new ServiceException(string.Format("could not load {0}", protocolPath));
			}

			var desc = new ProtocolDescription(schema, ns, contentType, constants);

			foreach (var constantId in constants.Keys) {
				var constant = constants[constantId];
				desc.Data[constantId] = constant;
			}

			var protocolDescriptions = xd.GetElementsByTagName("protocol_description");

			if (protocolDescriptions == null || protocolDescriptions.Count == 0)
				throw new ServiceException(string.Format("no 'protocol_description' element found in {0}", protocolPath));

			if (protocolDescriptions.Count > 1)
				throw new ServiceException(string.Format("too many 'protocol_description' elements found in {0}", protocolPath));

			var xeProtocolDescription = (XmlElement)protocolDescriptions[0];

			var fields = xeProtocolDescription.GetElementsByTagName("fields");

			if (fields == null || fields.Count == 0)
				throw new ServiceException(string.Format("no 'fields' element found in {0}/protocol_description", protocolPath));

			if (fields.Count > 1)
				throw new ServiceException(string.Format("too many 'fields' elements found in {0}/protocol_description", protocolPath));

			var xeFields = (XmlElement)fields[0];

			foreach (XmlElement xeField in xeFields.ChildNodes) {
				var field = GetField(protocolPath, xeField, desc.Fields, desc.Data, "fields");

				desc.Data[field.Id] = field;
				desc.Fields[field.Id] = field;
			}

			var dataSources = xeProtocolDescription.GetElementsByTagName("data_source");

			if (dataSources == null || dataSources.Count == 0)
				throw new ServiceException(string.Format("no 'data_source' element found in {0}/protocol_description", protocolPath));

			if (dataSources.Count > 1)
				throw new ServiceException(string.Format("too many 'data_source' elements found in {0}/protocol_description", protocolPath));

			var xeDataSource = (XmlElement)dataSources[0];

			var sqlCount = GetSQL(protocolPath, desc, xeDataSource, "count", "data_source");
			var sqlList = GetSQL(protocolPath, desc, xeDataSource, "list", "data_source");
			var sqlRecord = GetSQL(protocolPath, desc, xeDataSource, "record", "data_source");
			var sqlRecordList = GetSQL(protocolPath, desc, xeDataSource, "record_list", "data_source");
			var sqlRecordInfo = GetSQL(protocolPath, desc, xeDataSource, "record_info", "data_source");

			desc.DataSource = new DataSource(sqlCount, sqlList, sqlRecord, sqlRecordList, sqlRecordInfo);

			var collections = xeProtocolDescription.GetElementsByTagName("collections");

			if (collections.Count > 0) {
				var xeCollections = (XmlElement)collections[0];

				foreach (XmlElement xeCollection in xeCollections.ChildNodes) {
					var collection = GetCollection(protocolPath, desc, xeCollection);
					desc.Collections[collection.Id] = collection;
				}
			}

			var xeOutput = (XmlElement)xeProtocolDescription.GetElementsByTagName("output")[0];
			desc.Output.Template = xeOutput.GetAttribute("template");

			foreach (XmlElement xeNamedDataNode in xeOutput.ChildNodes) {
				switch (xeNamedDataNode.Name) {
					case "data_template": {
							var dt = GetDataTemplate(protocolPath, desc.Data, xeNamedDataNode, "output/data_template");

							desc.Data[dt.Id] = dt;
							desc.Output.DataTemplates[dt.Id] = dt;
							desc.OrderedDataKeys.Add(dt.Id);
						}
						break;
					case "conditional_data": {
							var cd = GetConditionalData(protocolPath, desc.Data, xeNamedDataNode, "output/conditional_data");

							desc.Data[cd.Id] = cd;
							desc.Output.ConditionalData[cd.Id] = cd;
							desc.OrderedDataKeys.Add(cd.Id);
						}
						break;
					case "data_translation": {
							var dt = GetDataTranslation(protocolPath, desc.Data, xeNamedDataNode, "output/data_translation");

							desc.Data[dt.Id] = dt;
							desc.Output.DataTranslations[dt.Id] = dt;
							desc.OrderedDataKeys.Add(dt.Id);
						}
						break;
					case "max_value": {
							var mv = GetMaxValue(protocolPath, desc.Data, xeNamedDataNode, "output/max_value");

							desc.Data[mv.Id] = mv;
							desc.Output.MaxValues[mv.Id] = mv;
							desc.OrderedDataKeys.Add(mv.Id);
						}
						break;
					case "min_value": {
							var mv = GetMinValue(protocolPath, desc.Data, xeNamedDataNode, "output/min_value");

							desc.Data[mv.Id] = mv;
							desc.Output.MinValues[mv.Id] = mv;
							desc.OrderedDataKeys.Add(mv.Id);
						}
						break;
				}
			}

			try {
				using (var sr = new StreamReader(string.Format("{0}\\{1}", PropertiesPath, desc.Output.Template))) {
					_templates[desc.Output.Template] = sr.ReadToEnd();
				}
			} catch {
				throw new ServiceException(string.Format("problem reading template {0}", desc.Output.Template));
			}

			return desc;
		}

		private static MinValue GetMinValue(string protocolPath, IDictionary<string, NamedData> data, XmlElement xeMinValueNode, string tag) {
			if (!xeMinValueNode.HasAttribute("id"))
				throw new ServiceException(string.Format("no 'id' attribute found in {0}/protocol_description/{1}[?]", protocolPath, tag));

			var id = xeMinValueNode.GetAttribute("id");

			if (!xeMinValueNode.HasAttribute("type"))
				throw new ServiceException(string.Format("no 'type' attribute found in {0}/protocol_description/{1}[?]", protocolPath, tag));

			var type = xeMinValueNode.GetAttribute("type");

			switch (type) {
				case "date":
				case "int":
				case "long":
				case "float":
				case "double":
					break;
				default:
					throw new ServiceException(string.Format("non-supported 'type' {0} found in {1}/protocol_description/{2}[?]", type, protocolPath, tag));
			}

			var sourcesNodes = xeMinValueNode.ChildNodes;

			if (sourcesNodes == null || sourcesNodes.Count == 0)
				throw new ServiceException(string.Format("no 'source' elements found in {0}", string.Format("no 'type' attribute found in {0}/protocol_description/{1}[?]", protocolPath, tag)));

			var sources = new List<NamedData>();

			foreach (XmlElement sourceNode in sourcesNodes) {
				sources.Add(data[sourceNode.InnerText]);
			}

			return new MinValue(id, sources, type);
		}

		private static MaxValue GetMaxValue(string protocolPath, IDictionary<string, NamedData> data, XmlElement xeMaxValueNode, string tag) {
			if (!xeMaxValueNode.HasAttribute("id"))
				throw new ServiceException(string.Format("no 'id' attribute found in {0}/protocol_description/{1}[?]", protocolPath, tag));

			var id = xeMaxValueNode.GetAttribute("id");

			if (!xeMaxValueNode.HasAttribute("type"))
				throw new ServiceException(string.Format("no 'type' attribute found in {0}/protocol_description/{1}[?]", protocolPath, tag));

			var type = xeMaxValueNode.GetAttribute("type");

			switch (type) {
				case "date":
				case "int":
				case "long":
				case "float":
				case "double":
					break;
				default:
					throw new ServiceException(string.Format("non-supported 'type' {0} found in {1}/protocol_description/{2}[?]", type, protocolPath, tag));
			}

			var sourcesNodes = xeMaxValueNode.ChildNodes;

			if (sourcesNodes == null || sourcesNodes.Count == 0)
				throw new ServiceException(string.Format("no 'source' elements found in {0}", string.Format("no 'type' attribute found in {0}/protocol_description/{1}[?]", protocolPath, tag)));

			var sources = new List<NamedData>();

			foreach (XmlElement sourceNode in sourcesNodes) {
				sources.Add(data[sourceNode.InnerText]);
			}

			return new MaxValue(id, sources, type);
		}

		private static DataTranslation GetDataTranslation(string protocolPath, IDictionary<string, NamedData> data, XmlElement xeDataTranslation, string tag) {
			if (!xeDataTranslation.HasAttribute("id"))
				throw new ServiceException(string.Format("no 'id' attribute found in {0}/protocol_description/{1}[?]", protocolPath, tag));

			var id = xeDataTranslation.GetAttribute("id");

			if (!xeDataTranslation.HasAttribute("source"))
				throw new ServiceException(string.Format("no 'source' attribute found in {0}/protocol_description/{1}[{2}]", protocolPath, tag, id));

			var source = xeDataTranslation.GetAttribute("source");

			if (!data.ContainsKey(source))
				throw new ServiceException(string.Format("source {0} not found in data", source));

			var dt = new DataTranslation(id, data[source]);

			foreach (XmlElement xeTranslation in xeDataTranslation.ChildNodes) {
				if (!xeTranslation.HasAttribute("source"))
					throw new ServiceException(string.Format("no 'source' attribute found in {0}/protocol_description/{1}[{2}]/translation[?]", protocolPath, tag, id));

				var trSource = OAIUtil.XmlEncode(xeTranslation.GetAttribute("source"));

				if (!xeTranslation.HasAttribute("target"))
					throw new ServiceException(string.Format("no 'target' attribute found in {0}/protocol_description/{1}[{2}]/translation[{3}]", protocolPath, tag, id, trSource));

				dt.Translations[trSource] = data[xeTranslation.GetAttribute("target")];
			}
			return dt;
		}

		private static ConditionalData GetConditionalData(string protocolPath, IDictionary<string, NamedData> data, XmlElement xeConditionalData, string tag) {
			if (!xeConditionalData.HasAttribute("id"))
				throw new ServiceException(string.Format("no 'id' attribute found in {0}/protocol_description/{1}[?]", protocolPath, tag));

			var id = xeConditionalData.GetAttribute("id");

			if (!xeConditionalData.HasAttribute("source"))
				throw new ServiceException(string.Format("no 'source' attribute found in {0}/protocol_description/{1}[{2}]", protocolPath, tag, id));

			var source = xeConditionalData.GetAttribute("source");

			if (!data.ContainsKey(source))
				throw new ServiceException(string.Format("source {0} not found in data", source));

			var cd = new ConditionalData(id, data[source]);

			var xeConditionalOutput = (XmlElement)xeConditionalData.ChildNodes[0];
			cd.Output.Template = xeConditionalOutput.GetAttribute("template");

			foreach (XmlElement xeDataTemplate in xeConditionalOutput.GetElementsByTagName("conditional_data_template")) {
				var dt = GetDataTemplate(protocolPath, data, xeDataTemplate, string.Format("{0}[{1}]/conditional_data_template", tag, id));
				cd.Output.DataTemplates[dt.Id] = dt;
			}

			using (var sr = new StreamReader(string.Format("{0}\\{1}", PropertiesPath, cd.Output.Template))) {
				_templates[cd.Output.Template] = sr.ReadToEnd();
			}
			return cd;
		}

		private static SQLStatement GetSQL(string protocolPath, ProtocolDescription desc, XmlElement xeDataSource, string tagName, string tagPath) {
			var tags = xeDataSource.GetElementsByTagName(tagName);

			if (tags == null || tags.Count == 0)
				throw new ServiceException(string.Format("no '{0}' element found in {1}/protocol_description/data_source", tagName, protocolPath));

			if (tags.Count > 1)
				throw new ServiceException(string.Format("too many '{0}' elements found in {1}/protocol_description/data_source", tagName, protocolPath));

			var xe = (XmlElement)tags[0];

			if (!xe.HasAttribute("name"))
				throw new ServiceException(string.Format("no 'name' attribute found in {0}/protocol_description/{1}/{2}", protocolPath, tagPath, tagName));

			if (!xe.HasAttribute("template"))
				throw new ServiceException(string.Format("no 'template' attribute found in {0}/protocol_description/{1}/{2}", protocolPath, tagPath, tagName));

			var sp = new SQLStatement(xe.GetAttribute("name"), xe.GetAttribute("template"));

			if (!_sql.ContainsKey(sp.Name)) {
				try {
					using (var sr = new StreamReader(string.Format("{0}\\{1}", PropertiesPath, sp.Template))) {
						_sql[sp.Name] = sr.ReadToEnd();
					}
				} catch {
					throw new ServiceException(string.Format("problem reading template {0}", sp.Template));
				}
			}

			if (xe.HasChildNodes) {
				var parameters = GetParameters(protocolPath, desc.Data, string.Format("{0}/{1}", tagPath, tagName), xe);

				for (var i = 0; i < parameters.Count; ++i) {
					if (!parameters.ContainsKey(i))
						throw new ServiceException(string.Format("parameter with index {0} is missing in {1}/protocol_description/{2}/{3}", i, protocolPath, tagPath, tagName));

					sp.Parameters.Add(parameters[i]);
				}
			}
			return sp;
		}

		private static Field GetField(string protocolPath, XmlElement xeField, IDictionary<string, Field> fields, IDictionary<string, NamedData> data, string tag) {
			if (!xeField.HasAttribute("id"))
				throw new ServiceException(string.Format("no 'id' attribute found in {0}/protocol_description/{1}/field[?]", protocolPath, tag));

			var id = xeField.GetAttribute("id");

			if (fields.ContainsKey(id))
				throw new ServiceException(string.Format("field {0} defined twice in {1}/protocol_description/{2}", id, protocolPath, tag));

			if (data.ContainsKey(id))
				throw new ServiceException(string.Format("named data {0} defined twice in {1}/protocol_description/{2}", id, protocolPath, tag));

			if (!xeField.HasAttribute("value"))
				throw new ServiceException(string.Format("no 'value' attribute found in {0}/protocol_description/{1}fields/field[{2}]", protocolPath, tag, id));

			var isIdentifier = false;

			if (xeField.HasAttribute("identifier") && xeField.GetAttribute("identifier") == "true") {
				isIdentifier = true;
			}

			var field = new Field(id, xeField.GetAttribute("value"), isIdentifier);

			if (xeField.HasAttribute("type"))
				field.Type = xeField.GetAttribute("type");
			return field;
		}

		private static Dictionary<int, Parameter> GetParameters(string protocolPath, IDictionary<string, NamedData> data, string tag, XmlNode xe) {
			var parameters = new Dictionary<int, Parameter>();

			foreach (XmlElement xeParameter in xe.ChildNodes) {
				if (!xeParameter.HasAttribute("index"))
					throw new ServiceException(
						string.Format("no 'index' attribute found in {0}/protocol_description/{1}/parameter[?]",
									  protocolPath, tag));

				var index = int.Parse(xeParameter.GetAttribute("index"));

				if (parameters.ContainsKey(index))
					throw new ServiceException(
						string.Format("parameter with index {0} defined twice in {1}/protocol_description/{2}", index,
									  protocolPath, tag));

				if (!xeParameter.HasAttribute("source"))
					throw new ServiceException(
						string.Format("no 'source' attribute found in {0}/protocol_description/data_source/{1}/parameter[{2}]",
									  protocolPath, tag, index));

				var source = xeParameter.GetAttribute("source");

				if (!data.ContainsKey(source))
					throw new ServiceException(string.Format("source {0} not found in data", source));
				parameters[index] = new Parameter(index, data[source]);
			}
			return parameters;
		}

		private static Collection GetCollection(string protocolPath, ProtocolDescription desc, XmlElement xeCollection) {
			var collId = xeCollection.GetAttribute("id");

			string delimiter = null;

			if (xeCollection.HasAttribute("delimiter"))
				delimiter = xeCollection.GetAttribute("delimiter");

			var dataSources = xeCollection.GetElementsByTagName("coll_data_source");

			if (dataSources == null || dataSources.Count == 0)
				throw new ServiceException(string.Format("no 'coll_data_source' element found in {0}/protocol_description/collections/collection[{1}]", protocolPath, collId));

			if (dataSources.Count > 1)
				throw new ServiceException(string.Format("too many 'coll_data_source' elements found in {0}/protocol_description/collections/collection[{1}]", protocolPath, collId));

			var xeDataSource = (XmlElement)dataSources[0];

			var sqlList = GetSQL(protocolPath, desc, xeDataSource, "list", string.Format("collections/collection[{0}]/coll_data_source", collId));

			var dsCollection = new DataSource(null, sqlList, null, null, null);

			if (!xeCollection.HasAttribute("multifetch"))
				throw new ServiceException(string.Format("no 'multifetch' attribute found in {0}/protocol_description/collections/collection[{1}]", protocolPath, collId));

			var multifetch = bool.Parse(xeCollection.GetAttribute("multifetch"));

			var zeroPadRowNumber = 0;

			if (xeCollection.HasAttribute("rownumberzeropadding"))
				zeroPadRowNumber = int.Parse(xeCollection.GetAttribute("rownumberzeropadding"));

			var maxRowNumber = 0;

			if (xeCollection.HasAttribute("maxrownumber"))
				maxRowNumber = int.Parse(xeCollection.GetAttribute("maxrownumber"));

			var collection = new Collection(collId, dsCollection, delimiter, multifetch, zeroPadRowNumber, maxRowNumber);

			var xeCollectionFields = (XmlElement)xeCollection.GetElementsByTagName("coll_fields")[0];

			foreach (XmlElement xeCollectionField in xeCollectionFields.ChildNodes) {
				var field = GetField(protocolPath, xeCollectionField, collection.Fields, collection.Data, string.Format("collections/collection[{0}]/coll_fields", collId));

				collection.Data[field.Id] = field;
				collection.Fields[field.Id] = field;
			}

			if (collection.ZeroPadRowNumber > 0)
			{
				collection.Data["ROW_NUMBER"] = new RowNumber();
			}

			var xeCollectionOutput = (XmlElement)xeCollection.GetElementsByTagName("coll_output")[0];
			collection.Output.Template = xeCollectionOutput.GetAttribute("template");

			foreach (XmlElement xeNamedDataNode in xeCollectionOutput.ChildNodes) {
				switch (xeNamedDataNode.Name) {
					case "coll_data_template": {
							var dt = GetDataTemplate(protocolPath, collection.Data, xeNamedDataNode, string.Format("collections/collection[{0}]/coll_output/coll_data_template", collId));
							collection.Output.DataTemplates[dt.Id] = dt;
							collection.Data[dt.Id] = dt;
							collection.OrderedDataKeys.Add(dt.Id);
						}
						break;
					case "coll_conditional_data": {
							var cd = GetConditionalData(protocolPath, collection.Data, xeNamedDataNode, string.Format("collections/collection[{0}]/coll_output/coll_conditional_data", collId));
							collection.Output.ConditionalData[cd.Id] = cd;
							collection.Data[cd.Id] = cd;
							collection.OrderedDataKeys.Add(cd.Id);
						}
						break;
					case "coll_data_translation": {
							var dt = GetDataTranslation(protocolPath, collection.Data, xeNamedDataNode, string.Format("collections/collection[{0}]/coll_output/coll_data_translation", collId));
							collection.Output.DataTranslations[dt.Id] = dt;
							collection.Data[dt.Id] = dt;
							collection.OrderedDataKeys.Add(dt.Id);
						}
						break;
					case "coll_max_value": {
							var mv = GetMaxValue(protocolPath, collection.Data, xeNamedDataNode, string.Format("collections/collection[{0}]/coll_output/coll_max_value", collId));

							collection.Data[mv.Id] = mv;
							collection.Output.MaxValues[mv.Id] = mv;
							collection.OrderedDataKeys.Add(mv.Id);
						}
						break;
					case "coll_min_value": {
							var mv = GetMinValue(protocolPath, collection.Data, xeNamedDataNode, string.Format("collections/collection[{0}]/coll_output/coll_min_value", collId));

							collection.Data[mv.Id] = mv;
							collection.Output.MinValues[mv.Id] = mv;
							collection.OrderedDataKeys.Add(mv.Id);
						}
						break;
				}
			}

			try {
				using (var sr = new StreamReader(string.Format("{0}\\{1}", PropertiesPath, collection.Output.Template))) {
					_templates[collection.Output.Template] = sr.ReadToEnd();
				}
			} catch {
				throw new ServiceException(string.Format("problem reading template {0}", collection.Output.Template));
			}

			var additionalOutputList = xeCollection.GetElementsByTagName("coll_additional_output");

			foreach (XmlElement xeAdditionalOutput in additionalOutputList) {
				string additionalDelimiter = null;

				if (xeAdditionalOutput.HasAttribute("delimiter"))
					additionalDelimiter = xeAdditionalOutput.GetAttribute("delimiter");

				var additionalOutput = new AdditionalOutput(xeAdditionalOutput.GetAttribute("id"),
															xeAdditionalOutput.GetAttribute("template"),
															additionalDelimiter);
				collection.AdditionalOutputs[additionalOutput.Id] = additionalOutput;

				foreach (var dataKey in collection.Data.Keys) {
					additionalOutput.Data[dataKey] = collection.Data[dataKey];
				}

				foreach (XmlElement xeNamedDataNode in xeAdditionalOutput.ChildNodes) {
					switch (xeNamedDataNode.Name) {
						case "coll_data_template": {
								var dt = GetDataTemplate(protocolPath, additionalOutput.Data, xeNamedDataNode, string.Format("collections/collection[{0}]/coll_output/coll_data_template", collId));
								additionalOutput.DataTemplates[dt.Id] = dt;
								additionalOutput.Data[dt.Id] = dt;
								additionalOutput.OrderedDataKeys.Add(dt.Id);
							}
							break;
						case "coll_conditional_data": {
								var cd = GetConditionalData(protocolPath, additionalOutput.Data, xeNamedDataNode, string.Format("collections/collection[{0}]/coll_output/coll_conditional_data", collId));
								additionalOutput.ConditionalData[cd.Id] = cd;
								additionalOutput.Data[cd.Id] = cd;
								additionalOutput.OrderedDataKeys.Add(cd.Id);
							}
							break;
						case "coll_data_translation": {
								var dt = GetDataTranslation(protocolPath, additionalOutput.Data, xeNamedDataNode, string.Format("collections/collection[{0}]/coll_output/coll_data_translation", collId));
								additionalOutput.DataTranslations[dt.Id] = dt;
								additionalOutput.Data[dt.Id] = dt;
								additionalOutput.OrderedDataKeys.Add(dt.Id);
							}
							break;
						case "coll_max_value": {
								var mv = GetMaxValue(protocolPath, additionalOutput.Data, xeNamedDataNode, string.Format("collections/collection[{0}]/coll_output/coll_max_value", collId));

								additionalOutput.Data[mv.Id] = mv;
								additionalOutput.MaxValues[mv.Id] = mv;
								additionalOutput.OrderedDataKeys.Add(mv.Id);
							}
							break;
						case "coll_min_value": {
								var mv = GetMinValue(protocolPath, additionalOutput.Data, xeNamedDataNode, string.Format("collections/collection[{0}]/coll_output/coll_min_value", collId));
								additionalOutput.Data[mv.Id] = mv;
								additionalOutput.MinValues[mv.Id] = mv;
								additionalOutput.OrderedDataKeys.Add(mv.Id);
							}
							break;
					}
				}

				try {
					using (var sr = new StreamReader(string.Format("{0}\\{1}", PropertiesPath, additionalOutput.Template))) {
						_templates[additionalOutput.Template] = sr.ReadToEnd();
					}
				} catch {
					throw new ServiceException(string.Format("problem reading template {0}", additionalOutput.Template));
				}
			}
			return collection;
		}

		private static DataTemplate GetDataTemplate(string protocolPath, IDictionary<string, NamedData> data, XmlElement xeDataTemplate, string tag) {
			if (!xeDataTemplate.HasAttribute("id"))
				throw new ServiceException(string.Format("no 'id' attribute found in {0}/protocol_description/{1}[?]", protocolPath, tag));

			var id = xeDataTemplate.GetAttribute("id");

			if (!xeDataTemplate.HasAttribute("value"))
				throw new ServiceException(string.Format("no 'value' attribute found in {0}/protocol_description/{1}[{2}]", protocolPath, tag, id));

			var urlEncode = false;

			if (xeDataTemplate.HasAttribute("urlEncode"))
				urlEncode = (xeDataTemplate.GetAttribute("urlEncode") == "true");

			var dt = new DataTemplate(id, xeDataTemplate.GetAttribute("value"), urlEncode);

			if (xeDataTemplate.HasChildNodes) {
				var parameters = GetParameters(protocolPath, data, tag + "[" + id + "]", xeDataTemplate);

				for (var i = 0; i < parameters.Count; ++i) {
					if (!parameters.ContainsKey(i))
						throw new ServiceException(string.Format("parameter with index {0} is missing in {1}/protocol_description/{2}[{3}]", i, protocolPath, tag, id));

					dt.Parameters.Add(parameters[i]);
				}
			}
			return dt;
		}

		private static string SettingsConfigFile {
			get { return "settings.xml"; }
		}

		public void Reload() {
			LoadConfiguration(true);
		}

		public bool SupportsMetadataPrefix(string metadataPrefix) {
			var service = GetService();
			return service.Descriptions.ContainsKey(metadataPrefix);
		}

		public Record GetRecord(string identifier, string metadataPrefix, DateTime buildDate) {
			var service = GetService();
			var description = service.Descriptions[metadataPrefix];
			ContentType = description.ContentType;

			var sqlRecordInfo = description.DataSource.RecordInfo;
			var header = GetRecordInfo(service, description, sqlRecordInfo, identifier);

			if (header == null)
				throw new IdDoesNotExistException(identifier);

			var record = new Record { Header = header };
			if (header.Status != "deleted") {
				var headers = new List<Header> {header};
				var multifetchData = CallMultiFetchCollectionSPs(service, description, headers);
				record.Metadata = GetRecordMetadata(service, header.Identifier, description, buildDate, multifetchData);
			}

			return record;
		}

		private Header GetRecordInfo(Service service, ProtocolDescription description, SQLStatement ss, string identifier) {
			var sql = new StringBuilder(_sql[ss.Name]);
			sql.Replace("${id}", ExtractIdentifier(identifier));

			foreach (var constantId in description.Constants.Keys) {
				var constant = description.Constants[constantId];
				sql.Replace(string.Format("${{{0}}}", constantId), constant.Value);
			}

			var conn = service.Connection;
			var dbConn = GetConnection(conn);

#if DEBUG
			var then = DateTime.Now;
			System.Diagnostics.Debug.WriteLine(string.Format("{0}: {1}", ss.Name, then));
#endif

			var cmd = GetCommand(service.Connection, sql.ToString(), dbConn);
			cmd.CommandTimeout = service.CommandTimeout;

			Header header = null;

			try {
				var reader = cmd.ExecuteReader();

				try {
					if (reader.Read()) {
						header = new Header(this, reader);
					}
				} finally {
					reader.Close();
				}
			} catch (Exception ex) {
				throw new ServiceException("SQL: " + sql, ex);
			} 

#if DEBUG
			var now = DateTime.Now;
			System.Diagnostics.Debug.WriteLine(string.Format("{0}: {1}ms", ss.Name, (now - then).TotalMilliseconds));
#endif

			return header;
		}

		public IEnumerable<string> GetIdentifyAdminEmails() {
			if (AdminEmails == null)
				AdminEmails = new List<string>();

			return AdminEmails;
		}

		public IEnumerable<string> GetCustomIdentifyDescriptions() {
			// TODO: configurate this somewhere
			// TODO: namespaces
			return new List<string>();
		}

		public string GetRepositoryName() {
			var service = GetService();
			return service.Name;
		}

		public IdentifiersList GetIdentifiersList(string from, string until, string set, string metadataPrefix) {
			if (set != null)
				throw new NoSetHierarchyException();

			var service = GetService();

			var description = service.Descriptions[metadataPrefix];
			ContentType = description.ContentType;
			var countSQL = description.DataSource.Count;
			var countListSPConds = CreateCountListSPConds(from, until);
			var count = CallCountSP(service, description, countSQL, countListSPConds);

			var listSQL = description.DataSource.List;

			ResumptionMap map = null;
			var list = CallListSP(service, description, listSQL, countListSPConds, 0, count, metadataPrefix, from, until, ref map);

			var result = new IdentifiersList {
				Headers = list,
				ResumptionMap = map
			};
			return result;
		}

		private Service GetService() {
			var serviceName = string.Format("{0}/{1}", Institution.ToLower(), Service.ToLower());

			if (!_services.ContainsKey(serviceName))
				throw new ServiceException(string.Format("Service {0} not configured.", serviceName));

			return _services[serviceName];
		}

		private List<Header> CallListSP(Service service, ProtocolDescription description, SQLStatement ss, string conds, int startIndex, int count, String metadataPrefix, String from, String until, ref ResumptionMap map) {
			var conn = service.Connection;
			var dbConn = GetConnection(conn);

			var list = new List<Header>();

			var pageSize = service.ResultBlock == 0 ? count : service.ResultBlock;
			var sql = new StringBuilder(_sql[ss.Name]);
			sql.Replace("${conds}", conds);

			foreach (var constantId in description.Constants.Keys) {
				var constant = description.Constants[constantId];
				sql.Replace(string.Format("${{{0}}}", constantId), constant.Value);
			}

			int pageSizeReal;
			var maxPosts = service.ResultBlock + startIndex;
			if (maxPosts > count) {
				pageSizeReal = count - startIndex;
			} else {
				pageSizeReal = pageSize;
			}

			switch (conn.Type) {
				case Connection.ConnectionType.SQL2005:
					sql.Replace("${pageSizeReal}", pageSizeReal.ToString());
					sql.Replace("${maxPosts}", maxPosts.ToString());
					break;
				case Connection.ConnectionType.MYSQL:
					sql.Replace("${startIndex}", startIndex.ToString());
					sql.Replace("${pageSize}", pageSize.ToString());
					break;
			}
#if DEBUG
			var then = DateTime.Now;
			System.Diagnostics.Debug.WriteLine(string.Format("{0}: {1}", ss.Name, then));
#endif

			var cmd = GetCommand(service.Connection, sql.ToString(), dbConn);
			cmd.CommandTimeout = service.CommandTimeout;

			try {
				var reader = cmd.ExecuteReader();

				try {
					while (reader.Read()) {
						list.Add(new Header(this, reader));
					}
				} finally {
					reader.Close();
				}
			} catch (Exception ex) {
				throw new ServiceException("SQL: " + sql, ex);
			} 

#if DEBUG
			var now = DateTime.Now;
			System.Diagnostics.Debug.WriteLine(string.Format("{0}: {1}ms", ss.Name, (now - then).TotalMilliseconds));
#endif

			map = count > startIndex + list.Count ? new ResumptionMap(count, list.Count + startIndex, metadataPrefix, from, until) : null;

			return list;
		}

		private Dictionary<string, Dictionary<string, string>> CallRecordListSP(Service service, ProtocolDescription description, SQLStatement ss, string conds, int startIndex, int count, ICollection<Header> headers) {
			var conn = service.Connection;
			var dbConn = GetConnection(conn);

			var recordsData = new Dictionary<string, Dictionary<string, string>>();

			if (headers.Count == 0)
				return recordsData;

			var fields = new StringBuilder();

			foreach (var field in description.Fields.Keys) {
				fields.AppendFormat("{0}{1} as {2}", fields.Length > 0 ? ", " : "", description.Fields[field].Value, field);
			}

			var pageSize = service.ResultBlock == 0 ? count : service.ResultBlock;
			var sql = new StringBuilder(_sql[ss.Name]);
			sql.Replace("${conds}", conds);
			sql.Replace("${fields}", fields.ToString());
			sql.Replace("${headers}", GetHeaderList(headers, service.IdentifierType));

			foreach (var constantId in description.Constants.Keys) {
				var constant = description.Constants[constantId];
				sql.Replace(string.Format("${{{0}}}", constantId), constant.Value);
			}

			int pageSizeReal;
			var maxPosts = service.ResultBlock + startIndex;
			if (maxPosts > count) {
				pageSizeReal = count - startIndex;
			} else {
				pageSizeReal = pageSize;
			}

			switch (conn.Type) {
				case Connection.ConnectionType.SQL2005:
					sql.Replace("${pageSizeReal}", pageSizeReal.ToString());
					sql.Replace("${maxPosts}", maxPosts.ToString());
					break;
				case Connection.ConnectionType.MYSQL:
					sql.Replace("${startIndex}", startIndex.ToString());
					sql.Replace("${pageSize}", pageSize.ToString());
					break;
			}
#if DEBUG
			var then = DateTime.Now;
			System.Diagnostics.Debug.WriteLine(string.Format("{0}: {1}", ss.Name, then));
#endif

			var cmd = GetCommand(service.Connection, sql.ToString(), dbConn);
			cmd.CommandTimeout = service.CommandTimeout;

			try {
				var reader = cmd.ExecuteReader();

				try {
					while (reader.Read()) {
						var values = new Dictionary<string, string>();
						var i = 0;
						foreach (var field in description.Fields.Keys) {
							var f = description.Fields[field];
							values[field] = OAIUtil.ReplaceDbNull(reader, i, "", f.Type);

							if (f.IsIdentifier)
								recordsData[values[field]] = values;
							++i;
						}
					}
				} finally {
					reader.Close();
				}
			} catch (Exception ex) {
				throw new ServiceException("SQL: " + sql, ex);
			}
#if DEBUG
			var now = DateTime.Now;
			System.Diagnostics.Debug.WriteLine(string.Format("{0}: {1}ms", ss.Name, (now - then).TotalMilliseconds));
#endif

			return recordsData;
		}

		private string GetHeaderList(IEnumerable<Header> headers, string type)
		{
			var sb = new StringBuilder();

			foreach (var header in headers)
			{
				var id = ExtractIdentifier(header.Identifier);

				if (sb.Length > 0)
				{
					sb.Append(", ");
				}

				switch (type)
				{
					case "varchar":
						sb.Append("'");
						sb.Append(id);
						sb.Append("'");
						break;
					default:
						sb.Append(id);
						break;
				}
			}

			return sb.ToString();
		}

		private static string CreateCountListSPConds(string from, string until) {
			if (string.IsNullOrEmpty(from) && string.IsNullOrEmpty(until))
				return "1 = 1";
			if (!string.IsNullOrEmpty(from) && string.IsNullOrEmpty(until))
				return string.Format("DateStamp > '{0}'", from);
			return string.IsNullOrEmpty(from)
					? string.Format("DateStamp < '{0}'", until)
					: string.Format("DateStamp > '{0}' AND DateStamp < '{1}'", from, until);
		}

		private static int CallCountSP(Service service, ProtocolDescription description, SQLStatement ss, string conds) {
			var dbConn = GetConnection(service.Connection);

			var sql = new StringBuilder(_sql[ss.Name]);
			sql.Replace("${conds}", conds);

			foreach (var constantId in description.Constants.Keys) {
				var constant = description.Constants[constantId];
				sql.Replace(string.Format("${{{0}}}", constantId), constant.Value);
			}

#if DEBUG
			var then = DateTime.Now;
			System.Diagnostics.Debug.WriteLine(string.Format("{0}: {1}", ss.Name, then));
#endif

			var cmd = GetCommand(service.Connection, sql.ToString(), dbConn);
			cmd.CommandTimeout = service.CommandTimeout;

			var count = 0;

			try {
				var reader = cmd.ExecuteReader();


				try {
					if (reader.Read()) {
						count = reader.GetInt32(0);
					}
				} finally {
					reader.Close();
				}
			} catch (Exception ex) {
				throw new ServiceException("SQL: " + sql, ex);
			} 
#if DEBUG
			var now = DateTime.Now;
			System.Diagnostics.Debug.WriteLine(string.Format("{0}: {1}ms", ss.Name, (now - then).TotalMilliseconds));
#endif

			return count;
		}

		private static IDbConnection GetConnection(Connection conn) {
			if (!_connectionCache.ContainsKey(conn.Id)) {
				switch (conn.Type)
				{
					case Connection.ConnectionType.MYSQL:
						_connectionCache[conn.Id] = new MySqlConnection(conn.ConnectionString);
						break;
					default:
						_connectionCache[conn.Id] = new OleDbConnection(conn.ConnectionString);
						break;
				}
			}

			var dbConn = _connectionCache[conn.Id];

			if (dbConn.State != ConnectionState.Open) {
				dbConn.Open();
			}

			return dbConn;
		}

		public IdentifiersList GetIdentifiersList(string resumptionToken) {
			var map = new ResumptionMap(resumptionToken);

			var service = GetService();

			if (!service.Descriptions.ContainsKey(map.MetadataPrefix))
				throw new BadResumptionTokenException();

			var description = service.Descriptions[map.MetadataPrefix];
			ContentType = description.ContentType;

			var count = map.CompleteListSize.Value;
			var countListSPConds = CreateCountListSPConds(map.From, map.Until);

			var listSQL = description.DataSource.List;

			var list = CallListSP(service, description, listSQL, countListSPConds, map.Cursor.Value, count, map.MetadataPrefix, map.From, map.Until, ref map);

			var result = new IdentifiersList {
				Headers = list,
				ResumptionMap = map
			};
			return result;
		}

		public List<MetadataFormat> GetAllMetadataFormats() {
			var service = GetService();

			var formats = new List<MetadataFormat>();
			foreach (var descKey in service.Descriptions.Keys) {
				var desc = service.Descriptions[descKey];
				var metadataFormat = new MetadataFormat { Label = descKey, SchemaURL = desc.Schema, NamespaceURI = desc.NameSpace };
				formats.Add(metadataFormat);
			}

			if (formats.Count == 0)
				throw new NoMetadataFormatsException();

			return formats;
		}

		public NameValueCollection GetQuery() {
			return Query;
		}

		public string GetBaseURL() {
			return BaseURL;
		}

		public string GetIdentifyEarliestDatestamp() {
			var service = GetService();
			return service.EarliestDatestamp;
		}

		public string GetDeletedRecord() {
			var service = GetService();
			return service.DeletedRecord;
		}

		public string GetIdentifyPrefix() {
			var service = GetService();
			return string.Format("{0}:{1}/", service.IdentifyPrefix, service.Id);
		}

		public string GetGranularity() {
			return Granularity;
		}

		public RecordsList GetRecordsList(string from, string until, string set, string metadataPrefix, DateTime buildDate) {
			if (set != null)
				throw new NoSetHierarchyException();

			var service = GetService();
			var description = service.Descriptions[metadataPrefix];
			ContentType = description.ContentType;
			var countSQL = description.DataSource.Count;
			var countListSPConds = CreateCountListSPConds(from, until);
			var count = CallCountSP(service, description, countSQL, countListSPConds);

			var listSQL = description.DataSource.List;

			ResumptionMap map = null;
			var headers = CallListSP(service, description, listSQL, countListSPConds, 0, count, metadataPrefix, from, until, ref map);

			var recordListSQL = description.DataSource.RecordList;

			var recordsData = CallRecordListSP(service, description, recordListSQL, countListSPConds, 0, count, headers);
			var multifetchData = CallMultiFetchCollectionSPs(service, description, headers);

			var records = new List<Record>();

			foreach (var header in headers) {
				var record = new Record { Header = header };
				if (header.Status != "deleted") {
					var identifier = ExtractIdentifier(header.Identifier);
					record.Metadata = GetRecordMetadata(service, description, buildDate, recordsData[identifier], identifier, multifetchData);
				}
				records.Add(record);
			}

			var result = new RecordsList {
				Records = records,
				ResumptionMap = map
			};

			if (result == null || result.Records.Count == 0)
				throw new NoRecordsMatchException();

			return result;
		}

		private IDictionary<string, IDictionary<string, IList<IDictionary<string, string>>>> CallMultiFetchCollectionSPs(Service service, ProtocolDescription description, IEnumerable<Header> headers)
		{
			var results = new Dictionary<string, IDictionary<string, IList<IDictionary<string, string>>>>();
			var dbConn = GetConnection(service.Connection);

			foreach (var collectionKey in description.Collections.Keys)
			{
				var collection = description.Collections[collectionKey];

				if (!collection.MultiFetch) continue;

				var headersList = GetHeaderList(headers, service.IdentifierType);

				if (headersList.Length > 0)
				{
					var collectionResults = GetMultiFetchCollectionValues(service, headersList, collection, dbConn);

					results[collectionKey] = collectionResults;
				}
				else
				{
					results[collectionKey] = new Dictionary<string, IList<IDictionary<string, string>>>();
				}
			}

			return results;
		}

		private static IDictionary<string, IList<IDictionary<string, string>>> GetMultiFetchCollectionValues(Service service, string headers, Collection collection, IDbConnection dbConn)
		{
			var collectionResults = new Dictionary<string, IList<IDictionary<string, string>>>();

			var fields = new StringBuilder();

			foreach (var field in collection.Fields.Keys) {
				fields.AppendFormat("{0}{1} as {2}", fields.Length > 0 ? ", " : "", collection.Fields[field].Value, field);
			}

			var sql = new StringBuilder(_sql[collection.DataSource.List.Name]);
			sql.Replace("${fields}", fields.ToString());
			sql.Replace("${headers}", headers);

#if DEBUG
			var then = DateTime.Now;
			System.Diagnostics.Debug.WriteLine(string.Format("{0}: {1}", collection.DataSource.List.Name, then));
#endif

			var cmd = GetCommand(service.Connection, sql.ToString(), dbConn);
			cmd.CommandTimeout = service.CommandTimeout;

#if DEBUG
			var rows = 0;
#endif
			try {
				var reader = cmd.ExecuteReader();


				try {
					while (reader.Read()) {
#if DEBUG
						++rows;
#endif
						var identifier = reader.GetValue(0).ToString();

						if (!collectionResults.ContainsKey(identifier)) {
							collectionResults[identifier] = new List<IDictionary<string, string>>();
						}

						var collValues = collectionResults[identifier];

						var collRowValues = new Dictionary<string, string>();

						if (collection.MaxRowNumber > 0 && collValues.Count + 1 > collection.MaxRowNumber)
							continue;

						if (collection.ZeroPadRowNumber > 0) {
							collRowValues["ROW_NUMBER"] = GetRowNumber(collValues.Count + 1, collection.ZeroPadRowNumber);
						}

						var i = 1;
						foreach (var field in collection.Fields.Keys) {
							var f = collection.Fields[field];
							collRowValues[field] = OAIUtil.ReplaceDbNull(reader, i, "", f.Type);
							++i;
						}

						collValues.Add(collRowValues);
					}
				} finally {
					reader.Close();
				}
			} catch (Exception ex) {
				throw new ServiceException("SQL: " + sql, ex);
			}

#if DEBUG
			var now = DateTime.Now;
			System.Diagnostics.Debug.WriteLine(string.Format("{0}: {1}ms ({2} rows)", collection.DataSource.List.Name, (now - then).TotalMilliseconds, rows));
#endif

			return collectionResults;
		}

		public string GetRecord(string protocol, string identifier) {
			var service = GetService();

			if (!service.Descriptions.ContainsKey(protocol))
				throw new ServiceException(string.Format("protocol '{0}' not supported.", protocol));

			var description = service.Descriptions[protocol];
			ContentType = description.ContentType;
			var buildDate = DateTime.Now;

			try {
				var header = new Header { Identifier = identifier };
				var headers = new List<Header> { header };
				var multifetchData = CallMultiFetchCollectionSPs(service, description, headers);
				return GetRecordMetadata(service, identifier, description, buildDate, multifetchData);
			} catch (IdDoesNotExistException) {
				throw new ServiceException(string.Format("Record with Id={0} does not exist", identifier));
			}
		}

		private string GetRecordMetadata(Service service, string id, ProtocolDescription description, DateTime buildDate, IDictionary<string, IDictionary<string, IList<IDictionary<string, string>>>> multifetchData) {
			var identifier = ExtractIdentifier(id);

			var template = new StringBuilder(_templates[description.Output.Template]);
			template.Replace("${BUILDDATE_YYYY-MM-DD}", buildDate.ToString("yyyy-MM-dd"));

			var fields = new StringBuilder();

			foreach (var field in description.Fields.Keys) {
				fields.AppendFormat("{0}{1} as {2}", fields.Length > 0 ? ", " : "", description.Fields[field].Value, field);
			}

			var dbConn = GetConnection(service.Connection);
			var sql = new StringBuilder(_sql[description.DataSource.Record.Name]);
			sql.Replace("${fields}", fields.ToString());
			sql.Replace("${id}", identifier);


			foreach (var constantId in description.Constants.Keys) {
				var constant = description.Constants[constantId];
				sql.Replace(string.Format("${{{0}}}", constantId), constant.Value);
			}

#if DEBUG
			var then = DateTime.Now;
			System.Diagnostics.Debug.WriteLine(string.Format("{0}: {1}", description.DataSource.Record.Name, then));
#endif

			var cmd = GetCommand(service.Connection, sql.ToString(), dbConn);
			cmd.CommandTimeout = service.CommandTimeout;

			var values = new Dictionary<string, string>();

			foreach (var constantId in description.Constants.Keys) {
				var constant = description.Constants[constantId];
				constant.GetValue(_templates, values);
			}

			try {
				var reader = cmd.ExecuteReader();

				try {
					if (reader.Read()) {
						var i = 0;
						foreach (var field in description.Fields.Keys) {
							var f = description.Fields[field];
							values[field] = OAIUtil.ReplaceDbNull(reader, i, "", f.Type);
							++i;
						}
					} else {
						throw new IdDoesNotExistException(id);
					}
				} finally {
					reader.Close();
				}
			} catch (Exception ex) {
				throw new ServiceException("SQL: " + sql, ex);
			} 

			foreach (var value in values.Keys) {
				template.Replace(string.Format("${{{0}}}", value), values[value]);
			}

			foreach (var namedDataKey in description.OrderedDataKeys) {
				var namedData = description.Data[namedDataKey];
				template.Replace(string.Format("${{{0}}}", namedDataKey), namedData.GetValue(_templates, values));
			}

			foreach (var collectionKey in description.Collections.Keys) {
				var collection = description.Collections[collectionKey];

				IList<IDictionary<string, string>> collValues = null;
				if (multifetchData.ContainsKey(collectionKey)) {
					var multiFetchCollectionData = multifetchData[collectionKey];

					if (multiFetchCollectionData.ContainsKey(identifier))
					{
						collValues = multiFetchCollectionData[identifier];
					}
				} else {
					collValues = GetCollectionValues(service, collection, values, dbConn);
				}

				if (collValues != null && collValues.Count > 0) {
					var collectionOutput = new StringBuilder();

					foreach (var collRowValues in collValues) {
						var collTemplate = new StringBuilder(_templates[collection.Output.Template]);

						foreach (var collValue in collRowValues.Keys) {
							collTemplate.Replace(string.Format("${{{0}}}", collValue), collRowValues[collValue]);
						}

						foreach (var namedDataKey in collection.OrderedDataKeys) {
							var namedData = collection.Data[namedDataKey];
							collTemplate.Replace(string.Format("${{{0}}}", namedDataKey), namedData.GetValue(_templates, collRowValues));
						}

						if (collection.Delimiter != null) {
							if (collectionOutput.Length > 0)
								collectionOutput.Append(collection.Delimiter);
						}

						collectionOutput.Append(collTemplate);
					}
					template.Replace(string.Format("${{{0}}}", collectionKey), collectionOutput.ToString());
				} else {
					template.Replace(string.Format("${{{0}}}", collectionKey), "");
				}

				foreach (var additionalOutputKey in collection.AdditionalOutputs.Keys) {
					if (collValues != null && collValues.Count > 0) {
						var additionalOutput = collection.AdditionalOutputs[additionalOutputKey];

						var collectionOutput = new StringBuilder();

						foreach (var collRowValues in collValues) {
							var additionalCollTemplate = new StringBuilder(_templates[additionalOutput.Template]);

							foreach (var collValue in collRowValues.Keys) {
								additionalCollTemplate.Replace(string.Format("${{{0}}}", collValue), collRowValues[collValue]);
							}

							foreach (var namedDataKey in additionalOutput.OrderedDataKeys) {
								var namedData = additionalOutput.Data[namedDataKey];
								additionalCollTemplate.Replace(string.Format("${{{0}}}", namedDataKey), namedData.GetValue(_templates, collRowValues));
							}

							if (additionalOutput.Delimiter != null) {
								if (collectionOutput.Length > 0)
									collectionOutput.Append(additionalOutput.Delimiter);
							}

							collectionOutput.Append(additionalCollTemplate);
						}
						template.Replace(string.Format("${{{0}:{1}}}", collectionKey, additionalOutputKey), collectionOutput.ToString());
					} else {
						template.Replace(string.Format("${{{0}:{1}}}", collectionKey, additionalOutputKey), "");
					}
				}
			}

#if DEBUG
			var now = DateTime.Now;
			System.Diagnostics.Debug.WriteLine(string.Format("{0}: {1}ms", description.DataSource.Record.Name, (now - then).TotalMilliseconds));
#endif

			return template.ToString();
		}

		private static string GetRecordMetadata(Service service, ProtocolDescription description, DateTime buildDate, Dictionary<string, string> values, String identifier, IDictionary<string, IDictionary<string, IList<IDictionary<string, string>>>> multifetchData) {
			var template = new StringBuilder(_templates[description.Output.Template]);
			template.Replace("${BUILDDATE_YYYY-MM-DD}", buildDate.ToString("yyyy-MM-dd"));

#if DEBUG
			var then = DateTime.Now;
			//System.Diagnostics.Debug.WriteLine(string.Format("{0}: {1}", description.DataSource.Record.Name, then));
#endif

			foreach (var constantId in description.Constants.Keys) {
				var constant = description.Constants[constantId];
				constant.GetValue(_templates, values);
			}

			var dbConn = GetConnection(service.Connection);

			foreach (var value in values.Keys) {
				template.Replace(string.Format("${{{0}}}", value), values[value]);
			}

			foreach (var namedDataKey in description.OrderedDataKeys) {
				var namedData = description.Data[namedDataKey];
				template.Replace(string.Format("${{{0}}}", namedDataKey), namedData.GetValue(_templates, values));
			}

			foreach (var collectionKey in description.Collections.Keys) {
				var collection = description.Collections[collectionKey];

				IList<IDictionary<string, string>> collValues = null;
				if (multifetchData.ContainsKey(collectionKey))
				{
					var multiFetchCollectionData = multifetchData[collectionKey];

					if (multiFetchCollectionData.ContainsKey(identifier))
					{
						collValues = multiFetchCollectionData[identifier];
					}
				}
				else
				{
					collValues = GetCollectionValues(service, collection, values, dbConn);
				}

				if (collValues != null && collValues.Count > 0) {
					var collectionOutput = new StringBuilder();

					foreach (var collRowValues in collValues) {
						var collTemplate = new StringBuilder(_templates[collection.Output.Template]);

						foreach (var collValue in collRowValues.Keys) {
							collTemplate.Replace(string.Format("${{{0}}}", collValue), collRowValues[collValue]);
						}

						foreach (var namedDataKey in collection.OrderedDataKeys) {
							var namedData = collection.Data[namedDataKey];

							collTemplate.Replace(string.Format("${{{0}}}", namedDataKey), namedData.GetValue(_templates, collRowValues));
						}

						if (collection.Delimiter != null) {
							if (collectionOutput.Length > 0)
								collectionOutput.Append(collection.Delimiter);
						}

						collectionOutput.Append(collTemplate);
					}
					template.Replace(string.Format("${{{0}}}", collectionKey), collectionOutput.ToString());
				} else {
					template.Replace(string.Format("${{{0}}}", collectionKey), "");
				}

				foreach (var additionalOutputKey in collection.AdditionalOutputs.Keys) {
					if (collValues != null && collValues.Count > 0) {
						var additionalOutput = collection.AdditionalOutputs[additionalOutputKey];

						var collectionOutput = new StringBuilder();

						// Deep copy needed for additional outputs, so that they don't interfer with each other
						// Keep primary collection values (collValues) though, instead of just taking fields (values)
						var additionalCollValues = new List<Dictionary<string, string>>();

						foreach (var collRowValues in collValues) {
							var additionalCollRowValues = new Dictionary<string, string>();

							foreach (var key in collRowValues.Keys) {
								additionalCollRowValues[key] = collRowValues[key];
							}

							additionalCollValues.Add(additionalCollRowValues);
						}

						foreach (var collRowValues in additionalCollValues) {
							var additionalCollTemplate = new StringBuilder(_templates[additionalOutput.Template]);

							foreach (var collValue in collRowValues.Keys) {
								additionalCollTemplate.Replace(string.Format("${{{0}}}", collValue), collRowValues[collValue]);
							}

							foreach (var namedDataKey in additionalOutput.OrderedDataKeys) {
								var namedData = additionalOutput.Data[namedDataKey];

								additionalCollTemplate.Replace(string.Format("${{{0}}}", namedDataKey), namedData.GetValue(_templates, collRowValues));
							}

							if (additionalOutput.Delimiter != null) {
								if (collectionOutput.Length > 0)
									collectionOutput.Append(additionalOutput.Delimiter);
							}

							collectionOutput.Append(additionalCollTemplate);
						}
						template.Replace(string.Format("${{{0}:{1}}}", collectionKey, additionalOutputKey), collectionOutput.ToString());
					} else {
						template.Replace(string.Format("${{{0}:{1}}}", collectionKey, additionalOutputKey), "");
					}
				}
			}

#if DEBUG
			var now = DateTime.Now;
			//System.Diagnostics.Debug.WriteLine(string.Format("{0}: {1}ms", description.DataSource.Record.Name, (now - then).TotalMilliseconds));
#endif

			return template.ToString();
		}

		private static IList<IDictionary<string, string>> GetCollectionValues(Service service, Collection collection, IDictionary<string, string> values, IDbConnection dbConn) {
			var fields = new StringBuilder();

			foreach (var field in collection.Fields.Keys) {
				fields.AppendFormat("{0}{1} as {2}", fields.Length > 0 ? ", " : "", collection.Fields[field].Value, field);
			}

			var sql = new StringBuilder(_sql[collection.DataSource.List.Name]);
			sql.Replace("${fields}", fields.ToString());

			foreach (var parameter in collection.DataSource.List.Parameters) {
				var source = parameter.Source;
				sql.Replace(string.Format("${{{0}}}", source.Id), source.GetValue(_templates, values));
			}

#if DEBUG
			var then = DateTime.Now;
			System.Diagnostics.Debug.WriteLine(string.Format("{0}: {1}", collection.DataSource.List.Name, then));
#endif

			var cmd = GetCommand(service.Connection, sql.ToString(), dbConn);
			cmd.CommandTimeout = service.CommandTimeout;

			IList<IDictionary<string, string>> collValues = new List<IDictionary<string, string>>();

			try {
				var reader = cmd.ExecuteReader();
				var row = 1;

				try {
					while (reader.Read()) {
						if (collection.MaxRowNumber > 0 && row >= collection.MaxRowNumber)
							break;

						var collRowValues = new Dictionary<string, string>();
						if (collection.ZeroPadRowNumber > 0)
						{
							collRowValues["ROW_NUMBER"] = GetRowNumber(row, collection.ZeroPadRowNumber);
						}
						var i = 0;
						foreach (var field in collection.Fields.Keys) {
							var f = collection.Fields[field];
							collRowValues[field] = OAIUtil.ReplaceDbNull(reader, i, "", f.Type);
							++i;
						}
						collValues.Add(collRowValues);
						++row;
					}
				} finally {
					reader.Close();
				}
			} catch (Exception ex) {
				throw new ServiceException("SQL: " + sql, ex);
			}

#if DEBUG
			var now = DateTime.Now;
			System.Diagnostics.Debug.WriteLine(string.Format("{0}: {1}ms", collection.DataSource.List.Name, (now - then).TotalMilliseconds));
#endif

			return collValues;
		}

		private static string GetRowNumber(int row, int number)
		{
			return row.ToString(string.Format("D{0}", number));
		}

		private static IDbCommand GetCommand(Connection connection, string sql, IDbConnection dbConn)
		{
			switch (connection.Type)
			{
				case Connection.ConnectionType.MYSQL:
					return new MySqlCommand(sql, (MySqlConnection)dbConn);
				default:
					return new OleDbCommand(sql, (OleDbConnection)dbConn);
			}
		}

		private string ExtractIdentifier(string identifier) {
			var id = identifier.Replace(string.Format("oai:{0}", GetIdentifyPrefix()), "");
			if (id.Contains("oai:") || id.Contains("/"))
				throw new IdDoesNotExistException(identifier);

			return id;
		}

		public RecordsList GetRecordsList(string resumptionToken, DateTime buildDate) {
			var map = new ResumptionMap(resumptionToken);

			var service = GetService();

			if (!service.Descriptions.ContainsKey(map.MetadataPrefix))
				throw new BadResumptionTokenException();

			var description = service.Descriptions[map.MetadataPrefix];
			ContentType = description.ContentType;
			var count = map.CompleteListSize.Value;
			var countListSPConds = CreateCountListSPConds(map.From, map.Until);

			var listSQL = description.DataSource.List;

			var cursor = map.Cursor.Value;

			var headers = CallListSP(service, description, listSQL, countListSPConds, cursor, count, map.MetadataPrefix, map.From, map.Until, ref map);

			var recordListSQL = description.DataSource.RecordList;

			var recordsData = CallRecordListSP(service, description, recordListSQL, countListSPConds, cursor, count, headers);
			var multifetchData = CallMultiFetchCollectionSPs(service, description, headers);

			var records = new List<Record>();

			foreach (var header in headers) {
				var record = new Record { Header = header };
				if (header.Status != "deleted") {
					var identifier = ExtractIdentifier(header.Identifier);
					record.Metadata = GetRecordMetadata(service, description, buildDate, recordsData[identifier], identifier, multifetchData);
				}
				records.Add(record);
			}

			var result = new RecordsList {
				Records = records,
				ResumptionMap = map
			};

			if (result == null || result.Records.Count == 0)
				throw new NoRecordsMatchException();

			return result;
		}

		public SetsList ListSets() {
			throw new NoSetHierarchyException();
		}

		public SetsList ListSets(string resumptionToken) {
			throw new NoSetHierarchyException();
		}
	}
}