﻿using System;
using System.Collections.Generic;
using System.Data.Common;
using System.Data.OleDb;
using System.Data.SqlClient;
using System.Diagnostics;
using System.IO;
using System.Net;
using System.Threading;
using System.Xml;
using MySql.Data.MySqlClient;
using OAI.Util;
using ServiceCore;
using ServiceCore.ServiceComponents;

namespace ServiceTester {
	class Program {
		static void Main(string[] args) {
			string servicesConfiguration;
			string servicesHttpRoot;

			var testConfigurationPath = @"..\..\testconfiguration.xml";

			if (args.Length > 0)
				testConfigurationPath = args[0];

			GetConfiguration(testConfigurationPath, out servicesConfiguration, out servicesHttpRoot);

			PropertiesPath = servicesConfiguration;
			HttpRoot = servicesHttpRoot;

			try
			{
				Trace.Listeners.Add(new TextWriterTraceListener(Console.Out));

				var start = DateTime.Now;

				LoadConfiguration();

				var end = DateTime.Now;
				Trace.WriteLine(string.Format("  OK: ServiceCore settings loaded ({0}ms)", (end - start).TotalMilliseconds));

				TestDbConnections();
				TestReload();

				TestHarvest("arkitekturmuseet/arkdok", "ksamsok-rdf");

				/*
				TestIdentify();
				TestListMetadataFormats();
				TestListIdentifiers();
				TestListSets();
				TestListRecords();*/

			}
			catch (Exception ex)
			{
				Trace.WriteLine(string.Format("FAIL: ServiceCore settings not loaded, reason {0}", ex.Message));
				Debug.WriteLine(ex.ToString());
			}

			Console.WriteLine("Press any key to quit");
			Console.ReadLine();
		}

		private static void TestHarvestWithResumptionToken(string serviceKey, string descKey, string token) {
			if (token == "")
				return;

			var serviceUrl = string.Format("{0}/{1}/", HttpRoot, serviceKey);
			var url = string.Format("{0}?verb=ListRecords&resumptionToken={1}", serviceUrl, token);

			var start = DateTime.Now;

			try {
				var request = (HttpWebRequest)WebRequest.Create(url);
				var service = _services[serviceKey];
				if (service.CommandTimeout > 0) {
					request.Timeout = service.CommandTimeout * 1000 + 5000;
				} else {
					request.Timeout = 3600000;
				}
				var response = request.GetResponse();

				var responseStream = response.GetResponseStream();
				var sr = new StreamReader(responseStream);

				var xd = new XmlDocument();
				xd.LoadXml(sr.ReadToEnd());

				response.Close();

				var listIdentifiers = (XmlElement)xd.GetElementsByTagName("ListRecords")[0];
				var identifier = listIdentifiers.GetElementsByTagName("identifier")[0].InnerText;

				if (!identifier.StartsWith(string.Format("oai:{0}:", service.IdentifyPrefix)))
					throw new Exception("identifier mismatch");

				if (xd.OuterXml.Contains("${"))
					throw new Exception("${x} found in TestHarvest");

				var end = DateTime.Now;
				Trace.WriteLine(string.Format("  OK: TestHarvest for service {0}, prefix {1}, token {2} ({3}ms)", serviceKey, descKey, token, (end - start).TotalMilliseconds));

				var resumptionTokens = listIdentifiers.GetElementsByTagName("resumptionToken");

				if (resumptionTokens.Count > 0)
					TestHarvestWithResumptionToken(serviceKey, descKey, resumptionTokens[0].InnerText);

			} catch (Exception ex) {
				Trace.WriteLine(string.Format("FAIL: TestHarvest for service {0}, prefix {1}, token {2} reason {3} [see: {4}]", serviceKey, descKey, token, ex.Message, url));
				Debug.WriteLine(ex.ToString());
			}
		}

		private static void TestHarvest(string serviceKey, string descKey)
		{
			var service = _services[serviceKey];
			var serviceUrl = string.Format("{0}/{1}/", HttpRoot, serviceKey);
			var url = string.Format("{0}?verb=ListRecords&metadataPrefix={1}", serviceUrl, descKey);

			var start = DateTime.Now;
			Trace.WriteLine(string.Format("STRT: TestHarvest for service {0}, prefix {1}", serviceKey, descKey));

			try {
				var request = (HttpWebRequest)WebRequest.Create(url);

				if (service.CommandTimeout > 0) {
					request.Timeout = service.CommandTimeout * 1000 + 5000;
				} else {
					request.Timeout = 3600000;
				}
				var response = request.GetResponse();

				var responseStream = response.GetResponseStream();
				var sr = new StreamReader(responseStream);

				var xd = new XmlDocument();
				xd.LoadXml(sr.ReadToEnd());

				response.Close();

				var listIdentifiers = (XmlElement)xd.GetElementsByTagName("ListRecords")[0];
				var identifier = listIdentifiers.GetElementsByTagName("identifier")[0].InnerText;

				if (!identifier.StartsWith(string.Format("oai:{0}:", service.IdentifyPrefix)))
					throw new Exception("identifier mismatch: " + identifier);

				if (xd.OuterXml.Contains("${"))
					throw new Exception("${x} found in TestHarvest");

				var end = DateTime.Now;
				Trace.WriteLine(string.Format("  OK: TestHarvest first for service {0}, prefix {1} ({2}ms)", serviceKey, descKey, (end - start).TotalMilliseconds));

				var resumptionTokens = listIdentifiers.GetElementsByTagName("resumptionToken");

				if (resumptionTokens.Count > 0)
					TestHarvestWithResumptionToken(serviceKey, descKey, resumptionTokens[0].InnerText);

				end = DateTime.Now;
				Trace.WriteLine(string.Format("  OK: TestHarvest finished for service {0}, prefix {1} ({2}ms)", serviceKey, descKey, (end - start).TotalMilliseconds));
			} catch (Exception ex) {
				Trace.WriteLine(string.Format("FAIL: TestHarvest for service {0}, prefix {1}, reason {2} [see: {3}]", serviceKey, descKey, ex.Message, url));
				Debug.WriteLine(ex.ToString());
			}
		}

		private static void TestReload()
		{
			foreach (var serviceKey in _services.Keys) {
				var serviceUrl = string.Format("{0}/{1}/", HttpRoot, serviceKey);
				var url = string.Format("{0}?verb=Reload", serviceUrl);

				var start = DateTime.Now;

				try {
					var request = (HttpWebRequest)WebRequest.Create(url);

					var service = _services[serviceKey];

					if (service.CommandTimeout > 0)
					{
						request.Timeout = service.CommandTimeout*1000 + 5000;
					}
					else
					{
						request.Timeout = 3600000;
					}

					var response = request.GetResponse();

					var responseStream = response.GetResponseStream();
					var sr = new StreamReader(responseStream);

					var xd = new XmlDocument();
					xd.LoadXml(sr.ReadToEnd());

					response.Close();

					var end = DateTime.Now;
					Trace.WriteLine(string.Format("  OK: Reload verb ({0}ms)", (end - start).TotalMilliseconds));
				} catch (Exception ex) {
					Trace.WriteLine(string.Format("FAIL: Reload verb reason {0} [see: {1}]", ex.Message, url));
					Debug.WriteLine(ex.ToString());
				}

				break;
			}
		}

		private static void TestListRecords()
		{
			foreach (var serviceKey in _services.Keys) {
				var service = _services[serviceKey];

				foreach (var descKey in service.Descriptions.Keys) {
					var serviceUrl = string.Format("{0}/{1}/", HttpRoot, serviceKey);
					var url = string.Format("{0}?verb=ListRecords&metadataPrefix={1}", serviceUrl, descKey);

					var start = DateTime.Now;

					try {
						var request = (HttpWebRequest)WebRequest.Create(url);

						if (service.CommandTimeout > 0) {
							request.Timeout = service.CommandTimeout * 1000 + 5000;
						} else {
							request.Timeout = 3600000;
						}
						var response = request.GetResponse();

						var responseStream = response.GetResponseStream();
						var sr = new StreamReader(responseStream);

						var xd = new XmlDocument();
						xd.LoadXml(sr.ReadToEnd());

						response.Close();

						var listIdentifiers = (XmlElement)xd.GetElementsByTagName("ListRecords")[0];
						var identifier = listIdentifiers.GetElementsByTagName("identifier")[0].InnerText;

						if (!identifier.StartsWith(string.Format("oai:{0}:", service.IdentifyPrefix)))
							throw new Exception("identifier mismatch");

						if (xd.OuterXml.Contains("${"))
							throw new Exception("${x} found in ListRecords");

						var end = DateTime.Now;
						Trace.WriteLine(string.Format("  OK: ListRecords verb for service {0}, prefix {1} ({2}ms)", serviceKey, descKey, (end - start).TotalMilliseconds));

						var resumptionTokens = listIdentifiers.GetElementsByTagName("resumptionToken");

						if (resumptionTokens.Count > 0)
							TestListRecordsWithResumptionToken(serviceKey, descKey, resumptionTokens[0].InnerText);

						TestListRecordsWithFromUntil(serviceKey, descKey);
					} catch (Exception ex) {
						Trace.WriteLine(string.Format("FAIL: ListRecords verb for service {0}, prefix {1}, reason {2} [see: {3}]", serviceKey, descKey, ex.Message, url));
						Debug.WriteLine(ex.ToString());
					}
				}
			}
		}

		private static void TestListRecordsWithFromUntil(string serviceKey, string descKey)
		{
			var serviceUrl = string.Format("{0}/{1}/", HttpRoot, serviceKey);
			var url = string.Format("{0}?verb=ListRecords&from={1}&until={2}&metadataPrefix={3}", serviceUrl, "2001-09-11", "2007-08-14", descKey);

			var start = DateTime.Now;

			try {
				var request = (HttpWebRequest)WebRequest.Create(url);
				Service service = _services[serviceKey];
				if (service.CommandTimeout > 0) {
					request.Timeout = service.CommandTimeout * 1000 + 5000;
				} else {
					request.Timeout = 3600000;
				}
				var response = request.GetResponse();

				var responseStream = response.GetResponseStream();
				var sr = new StreamReader(responseStream);

				var xd = new XmlDocument();
				xd.LoadXml(sr.ReadToEnd());

				response.Close();

				var end = DateTime.Now;
				Trace.WriteLine(string.Format("  OK: ListRecords verb with from/until for service {0}, prefix {1} ({2}ms)", serviceKey, descKey, (end - start).TotalMilliseconds));
			} catch (Exception ex) {
				Trace.WriteLine(string.Format("FAIL: ListRecords verb with from/until for service {0}, prefix {1}, reason {2} [see: {3}]", serviceKey, descKey, ex.Message, url));
				Debug.WriteLine(ex.ToString());
			}
		}

		private static void TestListRecordsWithResumptionToken(string serviceKey, string descKey, string token)
		{
			var serviceUrl = string.Format("{0}/{1}/", HttpRoot, serviceKey);
			var url = string.Format("{0}?verb=ListRecords&resumptionToken={1}", serviceUrl, token);

			var start = DateTime.Now;

			try {
				var request = (HttpWebRequest)WebRequest.Create(url);
				var service = _services[serviceKey];
				if (service.CommandTimeout > 0) {
					request.Timeout = service.CommandTimeout * 1000 + 5000;
				} else {
					request.Timeout = 3600000;
				}
				var response = request.GetResponse();

				var responseStream = response.GetResponseStream();
				var sr = new StreamReader(responseStream);

				var xd = new XmlDocument();
				xd.LoadXml(sr.ReadToEnd());

				response.Close();

				var listIdentifiers = (XmlElement)xd.GetElementsByTagName("ListRecords")[0];
				var identifier = listIdentifiers.GetElementsByTagName("identifier")[0].InnerText;

				if (!identifier.StartsWith(string.Format("oai:{0}:", service.IdentifyPrefix)))
					throw new Exception("identifier mismatch");

				var end = DateTime.Now;
				Trace.WriteLine(string.Format("  OK: ListRecords verb with resumptionToken for service {0}, prefix {1} ({2}ms)", serviceKey, descKey, (end - start).TotalMilliseconds));
			} catch (Exception ex) {
				Trace.WriteLine(string.Format("FAIL: ListRecords verb with resumptionToken for service {0}, prefix {1}, reason {2} [see: {3}]", serviceKey, descKey, ex.Message, url));
				Debug.WriteLine(ex.ToString());
			}
		}

		private static void TestListSets()
		{
			foreach (var serviceKey in _services.Keys) {
				var serviceUrl = string.Format("{0}/{1}/", HttpRoot, serviceKey);
				var url = string.Format("{0}?verb=ListSets", serviceUrl);

				var start = DateTime.Now;

				try {
					var request = (HttpWebRequest)WebRequest.Create(url);
					var service = _services[serviceKey];
					if (service.CommandTimeout > 0) {
						request.Timeout = service.CommandTimeout * 1000 + 5000;
					} else {
						request.Timeout = 3600000;
					}
					var response = request.GetResponse();

					var responseStream = response.GetResponseStream();
					var sr = new StreamReader(responseStream);

					var xd = new XmlDocument();
					xd.LoadXml(sr.ReadToEnd());

					response.Close();

					var end = DateTime.Now;
					Trace.WriteLine(string.Format("  OK: ListSets verb for service {0} ({1}ms)", serviceKey, (end - start).TotalMilliseconds));

				} catch (Exception ex) {
					Trace.WriteLine(string.Format("FAIL: ListSets verb for service {0}, reason {1} [see: {2}]", serviceKey, ex.Message, url));
					Debug.WriteLine(ex.ToString());
				}
			}
		}

		private static void TestListIdentifiers()
		{
			foreach (var serviceKey in _services.Keys) {
				var service = _services[serviceKey];

				foreach (var descKey in service.Descriptions.Keys) {
					var serviceUrl = string.Format("{0}/{1}/", HttpRoot, serviceKey);
					var url = string.Format("{0}?verb=ListIdentifiers&metadataPrefix={1}", serviceUrl, descKey);

					var start = DateTime.Now;

					try {
						var request = (HttpWebRequest)WebRequest.Create(url);
						if (service.CommandTimeout > 0) {
							request.Timeout = service.CommandTimeout * 1000 + 5000;
						} else {
							request.Timeout = 3600000;
						}
						var response = request.GetResponse();

						var responseStream = response.GetResponseStream();
						var sr = new StreamReader(responseStream);

						var xd = new XmlDocument();
						xd.LoadXml(sr.ReadToEnd());

						response.Close();
						var listIdentifiers = (XmlElement)xd.GetElementsByTagName("ListIdentifiers")[0];
						var identifier = listIdentifiers.GetElementsByTagName("identifier")[0].InnerText;

						if (!identifier.StartsWith(string.Format("oai:{0}:", service.IdentifyPrefix)))
							throw new Exception("identifier mismatch");

						var end = DateTime.Now;

						Trace.WriteLine(string.Format("  OK: ListIdentifiers verb for service {0}, prefix {1} ({2}ms)", serviceKey, descKey, (end - start).TotalMilliseconds));

						var resumptionTokens = listIdentifiers.GetElementsByTagName("resumptionToken");

						if (resumptionTokens.Count > 0)
							TestListIdentifiersWithResumptionToken(serviceKey, descKey, resumptionTokens[0].InnerText);

						TestGetRecord(serviceKey, descKey, identifier);
						TestGetRecordDirect(serviceKey, descKey, identifier);
					} catch (Exception ex) {
						Trace.WriteLine(string.Format("FAIL: ListIdentifiers verb for service {0}, prefix {1}, reason {2} [see: {3}]", serviceKey, descKey, ex.Message, url));
						Debug.WriteLine(ex.ToString());
					}
				}
			}
		}

		private static void TestGetRecord(string serviceKey, string descKey, string identifier) {
			var serviceUrl = string.Format("{0}/{1}/", HttpRoot, serviceKey);
			var url = string.Format("{0}?verb=GetRecord&metadataPrefix={1}&identifier={2}", serviceUrl, descKey, identifier);

			var start = DateTime.Now;

			try {
				var request = (HttpWebRequest)WebRequest.Create(url);
				var service = _services[serviceKey];
				if (service.CommandTimeout > 0) {
					request.Timeout = service.CommandTimeout * 1000 + 5000;
				} else {
					request.Timeout = 3600000;
				}
				var response = request.GetResponse();

				var responseStream = response.GetResponseStream();
				var sr = new StreamReader(responseStream);

				var xd = new XmlDocument();
				xd.LoadXml(sr.ReadToEnd());

				response.Close();
				if (xd.OuterXml.Contains("${"))
					throw new Exception("${x} found in GetRecord");

				var end = DateTime.Now;
				Trace.WriteLine(string.Format("  OK: GetRecord for service {0}, prefix {1} ({2}ms)", serviceKey, descKey, (end - start).TotalMilliseconds));
			} catch (Exception ex) {
				Trace.WriteLine(string.Format("FAIL: GetRecord for service {0}, prefix {1}, reason {2} [see: {3}]", serviceKey, descKey, ex.Message, url));
				Debug.WriteLine(ex.ToString());
			}
		}

		private static void TestGetRecordDirect(string serviceKey, string descKey, string identifier) {
			var url = string.Format("{0}/{1}/{2}/{3}", HttpRoot, serviceKey, descKey, identifier.Substring(identifier.IndexOf('/') + 1));

			var start = DateTime.Now;

			try {
				var request = (HttpWebRequest)WebRequest.Create(url);
				var service = _services[serviceKey];
				if (service.CommandTimeout > 0) {
					request.Timeout = service.CommandTimeout * 1000 + 5000;
				} else {
					request.Timeout = 3600000;
				}
				var response = request.GetResponse();

				var responseStream = response.GetResponseStream();
				var sr = new StreamReader(responseStream);

				var xd = new XmlDocument();
				xd.LoadXml(sr.ReadToEnd());

				response.Close();

				if (xd.OuterXml.Contains("${"))
					throw new Exception("${x} found in GetRecordDirect");

				var end = DateTime.Now;
				Trace.WriteLine(string.Format("  OK: GetRecordDirect for service {0}, prefix {1} ({2}ms)", serviceKey, descKey, (end - start).TotalMilliseconds));
			} catch (Exception ex) {
				Trace.WriteLine(string.Format("FAIL: GetRecordDirect for service {0}, prefix {1}, reason {2} [see: {3}]", serviceKey, descKey, ex.Message, url));
				Debug.WriteLine(ex.ToString());
			}
		}

		private static void TestListIdentifiersWithResumptionToken(string serviceKey, string descKey, string token) {
			var serviceUrl = string.Format("{0}/{1}/", HttpRoot, serviceKey);
			var url = string.Format("{0}?verb=ListIdentifiers&resumptionToken={1}", serviceUrl, token);

			var start = DateTime.Now;

			try {
				var request = (HttpWebRequest)WebRequest.Create(url);
				var service = _services[serviceKey];
				if (service.CommandTimeout > 0) {
					request.Timeout = service.CommandTimeout * 1000 + 5000;
				} else {
					request.Timeout = 3600000;
				}
				var response = request.GetResponse();

				var responseStream = response.GetResponseStream();
				var sr = new StreamReader(responseStream);

				var xd = new XmlDocument();
				xd.LoadXml(sr.ReadToEnd());

				response.Close();

				var listIdentifiers = (XmlElement)xd.GetElementsByTagName("ListIdentifiers")[0];
				var identifier = listIdentifiers.GetElementsByTagName("identifier")[0].InnerText;

				if (!identifier.StartsWith(string.Format("oai:{0}:", service.IdentifyPrefix)))
					throw new Exception("identifier mismatch");

				var end = DateTime.Now;

				Trace.WriteLine(string.Format("  OK: ListIdentifiers verb with resumptionToken for service {0}, prefix {1} ({2}ms)", serviceKey, descKey, (end - start).TotalMilliseconds));
			} catch (Exception ex) {
				Trace.WriteLine(string.Format("FAIL: ListIdentifiers verb with resumptionToken for service {0}, prefix {1}, reason {2} [see: {3}]", serviceKey, descKey, ex.Message, url));
				Debug.WriteLine(ex.ToString());
			}
		}

		private static void TestListMetadataFormats()
		{
			foreach (var serviceKey in _services.Keys) {
				var service = _services[serviceKey];

				var serviceUrl = string.Format("{0}/{1}/", HttpRoot, serviceKey);
				var url = string.Format("{0}?verb=ListMetadataFormats", serviceUrl);

				var start = DateTime.Now;

				try
				{
					var request = (HttpWebRequest)WebRequest.Create(url);
					if (service.CommandTimeout > 0) {
						request.Timeout = service.CommandTimeout * 1000 + 5000;
					} else {
						request.Timeout = 3600000;
					}
					var response = request.GetResponse();

					var responseStream = response.GetResponseStream();
					var sr = new StreamReader(responseStream);

					var xd = new XmlDocument();
					xd.LoadXml(sr.ReadToEnd());

					response.Close();
					var listMetadataFormats = (XmlElement)xd.GetElementsByTagName("ListMetadataFormats")[0];

					var metadataFormats = listMetadataFormats.GetElementsByTagName("metadataFormat");

					if (metadataFormats.Count != service.Descriptions.Count)
						throw new Exception("metadataFormats mismatch");

					foreach (XmlElement metadataFormat in metadataFormats)
					{
						var prefix = metadataFormat.ChildNodes[0].InnerText;

						if (!service.Descriptions.ContainsKey(prefix))
							throw new Exception(string.Format("metadataFormats mismatch, prefix {0}", prefix));


						var desc = service.Descriptions[prefix];

						var schema = metadataFormat.ChildNodes[1].InnerText;

						if (desc.Schema != schema)
							throw new Exception(string.Format("metadataFormats mismatch, schema {0} for {1}", schema, prefix));

						var ns = metadataFormat.ChildNodes[2].InnerText;

						if (desc.NameSpace != ns)
							throw new Exception(string.Format("metadataFormats mismatch, namespace {0} for {1}", ns, prefix));
					}

					var end = DateTime.Now;
					Trace.WriteLine(string.Format("  OK: ListMetadataFormats verb for service {0} ({1}ms)", serviceKey, (end - start).TotalMilliseconds));
				} catch (Exception ex) {
					Trace.WriteLine(string.Format("FAIL: ListMetadataFormats verb for service {0}, reason {1} [see: {2}]", serviceKey, ex.Message, url));
					Debug.WriteLine(ex.ToString());
				}
			}
		}

		private static void TestIdentify()
		{
			foreach (var serviceKey in _services.Keys) {
				var service = _services[serviceKey];

				var serviceUrl = string.Format("{0}/{1}/", HttpRoot, serviceKey);
				var url = string.Format("{0}?verb=Identify", serviceUrl);

				var start = DateTime.Now;
				try
				{
					var request = (HttpWebRequest)WebRequest.Create(url);
					if (service.CommandTimeout > 0) {
						request.Timeout = service.CommandTimeout * 1000 + 5000;
					} else {
						request.Timeout = 3600000;
					}
					var response = request.GetResponse();

					var responseStream = response.GetResponseStream();
					var sr = new StreamReader(responseStream);

					var xd = new XmlDocument();
					xd.LoadXml(sr.ReadToEnd());

					response.Close();
					var identify = (XmlElement)xd.GetElementsByTagName("Identify")[0];
					var repositoryName = identify.GetElementsByTagName("repositoryName")[0].InnerText;

					if (service.Name != repositoryName)
						throw new Exception("Service name mismatch");

					var baseURL = identify.GetElementsByTagName("baseURL")[0].InnerText;

					if (baseURL != serviceUrl)
						throw new Exception("baseURL mismatch");

					var adminEmails = identify.GetElementsByTagName("adminEmail");

					if (adminEmails.Count != AdminEmails.Count)
						throw new Exception("adminEmail mismatch");

					foreach (XmlElement adminEmail in adminEmails)
					{
						if (!AdminEmails.Contains(adminEmail.InnerText))
							throw new Exception("adminEmail mismatch");
					}

					var earliestDatestamp = identify.GetElementsByTagName("earliestDatestamp")[0].InnerText;

					if (earliestDatestamp != service.EarliestDatestamp)
						throw new Exception("earliestDatestamp mismatch");

					var deletedRecord = identify.GetElementsByTagName("deletedRecord")[0].InnerText;

					if (deletedRecord != service.DeletedRecord)
						throw new Exception("deletedRecord mismatch");

					var granularity = identify.GetElementsByTagName("granularity")[0].InnerText;

					if (granularity != Granularity)
						throw new Exception("granularity mismatch");

					var end = DateTime.Now;
					Trace.WriteLine(string.Format("  OK: Identify verb for service {0} ({1}ms)", serviceKey, (end - start).TotalMilliseconds));
				}
				catch (Exception ex)
				{
					Trace.WriteLine(string.Format("FAIL: Identify verb for service {0}, reason {1} [see: {2}]", serviceKey, ex.Message, url));
					Debug.WriteLine(ex.ToString());
				}
			}
		}

		private static void TestDbConnections() {
			foreach (var connectionKey in _connections.Keys) {
				var connection = _connections[connectionKey];
				var connectionString = connection.ConnectionString;

				var start = DateTime.Now;

				var conn = GetConnection(connection);

				try
				{
					conn.Open();

					var end = DateTime.Now;
					Trace.WriteLine(string.Format("  OK: Connection {0} opened ({1}ms)", connectionKey, (end - start).TotalMilliseconds));
				}
				catch (Exception ex)
				{
					Trace.WriteLine(string.Format("FAIL: Connection {0} could not open, reason: {1} [{2}]", connectionKey, ex.Message, connectionString));
					Debug.WriteLine(ex.ToString());
				}
			}
		}

		private static DbConnection GetConnection(Connection conn) {
			switch (conn.Type) {
				case Connection.ConnectionType.MYSQL:
					return new MySqlConnection(conn.ConnectionString);
				default:
					return new OleDbConnection(conn.ConnectionString);
			}
		}

		private static void GetConfiguration(string testConfigurationPath, out string servicesConfiguration, out string servicesHttpRoot) {
			var doc = new XmlDocument();
			doc.Load(testConfigurationPath);

			var configuration = (XmlElement) doc.GetElementsByTagName("configuration")[0];
			servicesConfiguration = configuration.GetElementsByTagName("services_configuration_path")[0].InnerText;
			servicesHttpRoot = configuration.GetElementsByTagName("services_http_root")[0].InnerText;
		}

		private static readonly Dictionary<string, Connection> _connections = new Dictionary<string, Connection>();
		private static readonly Dictionary<string, Service> _services = new Dictionary<string, Service>();
		private static readonly Dictionary<string, string> _templates = new Dictionary<string, string>();
		private static readonly Dictionary<string, string> _sql = new Dictionary<string, string>();
		public static string Granularity { get; private set; }
		public static List<string> AdminEmails { get; private set; }
		public static string PropertiesPath { get; set; }
		public static string HttpRoot { get; set; }

		private static string SettingsConfigFile {
			get { return "settings.xml"; }
		}
	
		public static void LoadConfiguration()
		{
			_connections.Clear();
			_services.Clear();
			_templates.Clear();
			_sql.Clear();

			var xd = new XmlDocument();

			try
			{
				xd.Load(string.Format("{0}\\{1}", PropertiesPath, SettingsConfigFile));
			}
			catch (XmlException)
			{
				throw new ServiceException(string.Format("{0} does not have well-formed xml", SettingsConfigFile));
			}
			catch (Exception)
			{
				throw new ServiceException(string.Format("could not load {0}", SettingsConfigFile));
			}

			var settings = xd.GetElementsByTagName("settings");

			if (settings == null || settings.Count == 0)
				throw new ServiceException(string.Format("no 'settings' element found in {0}", SettingsConfigFile));

			if (settings.Count > 1)
				throw new ServiceException(string.Format("too many 'settings' elements found in {0}", SettingsConfigFile));

			var xeSettings = (XmlElement) settings[0];

			var granularity = xeSettings.GetElementsByTagName("granularity");

			if (granularity == null || granularity.Count == 0)
				throw new ServiceException(string.Format("no 'granularity' element found in {0}/settings", SettingsConfigFile));

			Granularity = granularity[0].InnerText;

			var adminEmails = xeSettings.GetElementsByTagName("adminEmail");

			if (adminEmails == null || adminEmails.Count == 0)
				throw new ServiceException(string.Format("no 'adminEmail' elements found in {0}/settings", SettingsConfigFile));

			AdminEmails = new List<string>();

			foreach (XmlElement xeAdminEmail in adminEmails)
			{
				AdminEmails.Add(xeAdminEmail.InnerText);
			}

			var serviceConfigurations = xeSettings.GetElementsByTagName("serviceConfiguration");

			if (serviceConfigurations == null || serviceConfigurations.Count == 0)
				throw new ServiceException(string.Format("no 'serviceConfiguration' elements found in {0}/settings",
				                                         SettingsConfigFile));

			foreach (XmlElement xeServiceConfiguration in serviceConfigurations)
			{
				LoadServices(xeServiceConfiguration.InnerText);
			}
		}

		private static void LoadServices(string configuration) {
			var xd = new XmlDocument();

			try {
				xd.Load(string.Format("{0}\\{1}", PropertiesPath, configuration));
			} catch (XmlException) {
				throw new ServiceException(string.Format("{0} does not have well-formed xml", configuration));
			} catch (Exception) {
				throw new ServiceException(string.Format("could not load {0}", configuration));
			}

			var serviceConfigurations = xd.GetElementsByTagName("service_configuration");

			if (serviceConfigurations == null || serviceConfigurations.Count == 0)
				throw new ServiceException(string.Format("no 'service_configuration' element found in {0}", configuration));

			if (serviceConfigurations.Count > 1)
				throw new ServiceException(string.Format("too many 'service_configuration' elements found in {0}", configuration));

			var xeServiceConfiguration = (XmlElement)serviceConfigurations[0];

			var connections = xeServiceConfiguration.GetElementsByTagName("connections");

			if (connections == null || connections.Count == 0)
				throw new ServiceException(string.Format("no 'connections' element found in {0}/service_configuration", configuration));

			if (connections.Count > 1)
				throw new ServiceException(string.Format("too many 'connections' elements found in {0}/service_configuration", configuration));

			var xeConnections = (XmlElement)connections[0];

			if (!xeConnections.HasChildNodes)
				throw new ServiceException(string.Format("no 'connection' elements found in {0}/service_configuration/connections", configuration));

			foreach (XmlElement xeConnection in xeConnections.ChildNodes) {
				if (!xeConnection.HasAttribute("id"))
					throw new ServiceException(string.Format("no 'id' attribute found in {0}/service_configuration/connections/connection[?]", configuration));

				var id = xeConnection.GetAttribute("id");

				if (_connections.ContainsKey(id))
					throw new ServiceException(string.Format("connection {0} defined twice", id));

				if (!xeConnection.HasAttribute("connectionString"))
					throw new ServiceException(string.Format("no 'connectionString' attribute found in {0}/service_configuration/connections/connection[{1}]", configuration, id));

				if (!xeConnection.HasAttribute("type"))
					throw new ServiceException(string.Format("no 'type' attribute found in {0}/service_configuration/connections/connection[{1}]", configuration, id));

				var connection = new Connection(
					id,
					xeConnection.GetAttribute("connectionString"),
					xeConnection.GetAttribute("type"));
				_connections[connection.Id] = connection;
			}

			var services = xeServiceConfiguration.GetElementsByTagName("services");

			if (services == null || services.Count == 0)
				throw new ServiceException(string.Format("no 'services' element found in {0}/service_configuration", configuration));

			if (services.Count > 1)
				throw new ServiceException(string.Format("too many 'services' elements found in {0}/service_configuration", configuration));

			var xeServices = (XmlElement)services[0];

			if (!xeServices.HasChildNodes)
				throw new ServiceException(string.Format("no 'service' elements found in {0}/service_configuration/services", configuration));

			foreach (XmlElement xeService in xeServices.ChildNodes) {
				if (!xeService.HasAttribute("id"))
					throw new ServiceException(string.Format("no 'id' attribute found in {0}/service_configuration/services/service[?]", configuration));

				var id = xeService.GetAttribute("id");

				if (!xeService.HasAttribute("name"))
					throw new ServiceException(string.Format("no 'name' attribute found in {0}/service_configuration/services/service[{1}]", configuration, id));

				var name = xeService.GetAttribute("name");

				if (!xeService.HasAttribute("institution"))
					throw new ServiceException(string.Format("no 'institution' attribute found in {0}/service_configuration/services/service[{1}]", configuration, id));

				var institution = xeService.GetAttribute("institution");

				if (_services.ContainsKey(string.Format("{0}/{1}", institution.ToLower(), id.ToLower())))
					throw new ServiceException(string.Format("service {0}/{1} defined twice", institution.ToLower(), id.ToLower()));

				if (!xeService.HasAttribute("connection"))
					throw new ServiceException(string.Format("no 'connection' attribute found in {0}/service_configuration/services/service[{1}]", configuration, id));

				var connection = _connections[xeService.GetAttribute("connection")];

				if (!xeService.HasAttribute("update_rate"))
					throw new ServiceException(string.Format("no 'update_rate' attribute found in {0}/service_configuration/services/service[{1}]", configuration, id));

				var updateRate = int.Parse(xeService.GetAttribute("update_rate"));

				if (!xeService.HasAttribute("result_block"))
					throw new ServiceException(string.Format("no 'result_block' attribute found in {0}/service_configuration/services/service[{1}]", configuration, id));

				var resultBlock = int.Parse(xeService.GetAttribute("result_block"));

				if (!xeService.HasAttribute("deleted_record"))
					throw new ServiceException(string.Format("no 'deleted_record' attribute found in {0}/service_configuration/services/service[{1}]", configuration, id));

				var deletedRecord = xeService.GetAttribute("deleted_record");

				if (!xeService.HasAttribute("earliest_datestamp"))
					throw new ServiceException(string.Format("no 'earliest_datestamp' attribute found in {0}/service_configuration/services/service[{1}]", configuration, id));

				var earliestDatestamp = xeService.GetAttribute("earliest_datestamp");


				if (!xeService.HasAttribute("identify_prefix"))
					throw new ServiceException(string.Format("no 'identify_prefix' attribute found in {0}/service_configuration/services/service[{1}]", configuration, id));

				var identifyPrefix = xeService.GetAttribute("identify_prefix");

				if (!xeService.HasChildNodes)
					throw new ServiceException(string.Format("no 'protocol' elements found in {0}/service_configuration/services/service[{1}]", configuration, id));

				var commandTimeout = 0;

				if (xeService.HasAttribute("command_timeout"))
					commandTimeout = int.Parse(xeService.GetAttribute("command_timeout"));

				if (!xeService.HasAttribute("identifier_type"))
					throw new ServiceException(string.Format("no 'identifier_type' attribute found in {0}/service_configuration/services/service[{1}]", configuration, id));

				var identifierType = xeService.GetAttribute("identifier_type");

				var service = new Service(id, name, institution, connection, updateRate, resultBlock, deletedRecord, earliestDatestamp, identifyPrefix, commandTimeout, identifierType);
				_services[string.Format("{0}/{1}", institution.ToLower(), id.ToLower())] = service;

				foreach (XmlElement xeProtocol in xeService.ChildNodes) {
					if (!xeProtocol.HasAttribute("id"))
						throw new ServiceException(string.Format("no 'id' attribute found in {0}/service_configuration/services/service[{1}]/protocols/protocol[?]", configuration, id));

					var protocolId = xeProtocol.GetAttribute("id");

					if (!xeProtocol.HasAttribute("configuration"))
						throw new ServiceException(string.Format("no 'configuration' attribute found in {0}/service_configuration/services/service[{1}]/protocols/protocol[{2}]", configuration, id, protocolId));

					var protocolConfig = xeProtocol.GetAttribute("configuration");

					if (!xeProtocol.HasAttribute("schema"))
						throw new ServiceException(string.Format("no 'schema' attribute found in {0}/service_configuration/services/service[{1}]/protocols/protocol[{2}]", configuration, id, protocolId));

					var schema = xeProtocol.GetAttribute("schema");

					if (!xeProtocol.HasAttribute("namespace"))
						throw new ServiceException(string.Format("no 'namespace' attribute found in {0}/service_configuration/services/service[{1}]/protocols/protocol[{2}]", configuration, id, protocolId));

					var ns = xeProtocol.GetAttribute("namespace");

					if (!xeProtocol.HasAttribute("content-type"))
						throw new ServiceException(string.Format("no 'content-type' attribute found in {0}/service_configuration/services/service[{1}]/protocols/protocol[{2}]", configuration, id, protocolId));

					var contentType = xeProtocol.GetAttribute("content-type");

					var constants = new Dictionary<string, Constant>();

					if (xeProtocol.HasChildNodes) {
						foreach (XmlElement xeConstant in xeProtocol.ChildNodes) {
							if (!xeConstant.HasAttribute("id"))
								throw new ServiceException(string.Format("no 'id' attribute found in {0}/service_configuration/services/service[{1}]/protocols/protocol[{2}]/constant[?]", configuration, id, protocolId));

							var constantId = xeConstant.GetAttribute("id");

							if (!xeConstant.HasAttribute("value"))
								throw new ServiceException(string.Format("no 'value' attribute found in {0}/service_configuration/services/service[{1}]/protocols/protocol[{2}]/constant[{3}]", configuration, id, protocolId, constantId));

							constants[constantId] = new Constant(constantId, xeConstant.GetAttribute("value"));
						}
					}

					service.Descriptions[protocolId] = GetProtocolDescription(protocolConfig, schema, ns, contentType, constants);
				}
			}
		}

		private static ProtocolDescription GetProtocolDescription(string protocolPath, string schema, string ns, string contentType, Dictionary<string, Constant> constants) {
			var xd = new XmlDocument();

			try {
				xd.Load(string.Format("{0}\\{1}", PropertiesPath, protocolPath));
			} catch (XmlException) {
				throw new ServiceException(string.Format("{0} does not have well-formed xml", protocolPath));
			} catch (Exception) {
				throw new ServiceException(string.Format("could not load {0}", protocolPath));
			}

			var desc = new ProtocolDescription(schema, ns, contentType, constants);

			foreach (var constantId in constants.Keys) {
				var constant = constants[constantId];
				desc.Data[constantId] = constant;
			}

			var protocolDescriptions = xd.GetElementsByTagName("protocol_description");

			if (protocolDescriptions == null || protocolDescriptions.Count == 0)
				throw new ServiceException(string.Format("no 'protocol_description' element found in {0}", protocolPath));

			if (protocolDescriptions.Count > 1)
				throw new ServiceException(string.Format("too many 'protocol_description' elements found in {0}", protocolPath));

			var xeProtocolDescription = (XmlElement)protocolDescriptions[0];

			var fields = xeProtocolDescription.GetElementsByTagName("fields");

			if (fields == null || fields.Count == 0)
				throw new ServiceException(string.Format("no 'fields' element found in {0}/protocol_description", protocolPath));

			if (fields.Count > 1)
				throw new ServiceException(string.Format("too many 'fields' elements found in {0}/protocol_description", protocolPath));

			var xeFields = (XmlElement)fields[0];

			foreach (XmlElement xeField in xeFields.ChildNodes) {
				var field = GetField(protocolPath, xeField, desc.Fields, desc.Data, "fields");

				desc.Data[field.Id] = field;
				desc.Fields[field.Id] = field;
			}

			var dataSources = xeProtocolDescription.GetElementsByTagName("data_source");

			if (dataSources == null || dataSources.Count == 0)
				throw new ServiceException(string.Format("no 'data_source' element found in {0}/protocol_description", protocolPath));

			if (dataSources.Count > 1)
				throw new ServiceException(string.Format("too many 'data_source' elements found in {0}/protocol_description", protocolPath));

			var xeDataSource = (XmlElement)dataSources[0];

			var sqlCount = GetSQL(protocolPath, desc, xeDataSource, "count", "data_source");
			var sqlList = GetSQL(protocolPath, desc, xeDataSource, "list", "data_source");
			var sqlRecord = GetSQL(protocolPath, desc, xeDataSource, "record", "data_source");
			var sqlRecordList = GetSQL(protocolPath, desc, xeDataSource, "record_list", "data_source");
			var sqlRecordInfo = GetSQL(protocolPath, desc, xeDataSource, "record_info", "data_source");

			desc.DataSource = new DataSource(sqlCount, sqlList, sqlRecord, sqlRecordList, sqlRecordInfo);

			var collections = xeProtocolDescription.GetElementsByTagName("collections");

			if (collections.Count > 0) {
				var xeCollections = (XmlElement)collections[0];

				foreach (XmlElement xeCollection in xeCollections.ChildNodes) {
					var collection = GetCollection(protocolPath, desc, xeCollection);
					desc.Collections[collection.Id] = collection;
				}
			}

			var xeOutput = (XmlElement)xeProtocolDescription.GetElementsByTagName("output")[0];
			desc.Output.Template = xeOutput.GetAttribute("template");

			foreach (XmlElement xeNamedDataNode in xeOutput.ChildNodes) {
				switch (xeNamedDataNode.Name) {
					case "data_template": {
							var dt = GetDataTemplate(protocolPath, desc.Data, xeNamedDataNode, "output/data_template");

							desc.Data[dt.Id] = dt;
							desc.Output.DataTemplates[dt.Id] = dt;
							desc.OrderedDataKeys.Add(dt.Id);
						}
						break;
					case "conditional_data": {
							var cd = GetConditionalData(protocolPath, desc.Data, xeNamedDataNode, "output/conditional_data");

							desc.Data[cd.Id] = cd;
							desc.Output.ConditionalData[cd.Id] = cd;
							desc.OrderedDataKeys.Add(cd.Id);
						}
						break;
					case "data_translation": {
							var dt = GetDataTranslation(protocolPath, desc.Data, xeNamedDataNode, "output/data_translation");

							desc.Data[dt.Id] = dt;
							desc.Output.DataTranslations[dt.Id] = dt;
							desc.OrderedDataKeys.Add(dt.Id);
						}
						break;
					case "max_value": {
							var mv = GetMaxValue(protocolPath, desc.Data, xeNamedDataNode, "output/max_value");

							desc.Data[mv.Id] = mv;
							desc.Output.MaxValues[mv.Id] = mv;
							desc.OrderedDataKeys.Add(mv.Id);
						}
						break;
					case "min_value": {
							var mv = GetMinValue(protocolPath, desc.Data, xeNamedDataNode, "output/min_value");

							desc.Data[mv.Id] = mv;
							desc.Output.MinValues[mv.Id] = mv;
							desc.OrderedDataKeys.Add(mv.Id);
						}
						break;
				}
			}

			try {
				using (var sr = new StreamReader(string.Format("{0}\\{1}", PropertiesPath, desc.Output.Template))) {
					_templates[desc.Output.Template] = sr.ReadToEnd();
				}
			} catch {
				throw new ServiceException(string.Format("problem reading template {0}", desc.Output.Template));
			}

			return desc;
		}

		private static Field GetField(string protocolPath, XmlElement xeField, IDictionary<string, Field> fields, IDictionary<string, NamedData> data, string tag) {
			if (!xeField.HasAttribute("id"))
				throw new ServiceException(string.Format("no 'id' attribute found in {0}/protocol_description/{1}/field[?]", protocolPath, tag));

			var id = xeField.GetAttribute("id");

			if (fields.ContainsKey(id))
				throw new ServiceException(string.Format("field {0} defined twice in {1}/protocol_description/{2}", id, protocolPath, tag));

			if (data.ContainsKey(id))
				throw new ServiceException(string.Format("named data {0} defined twice in {1}/protocol_description/{2}", id, protocolPath, tag));

			if (!xeField.HasAttribute("value"))
				throw new ServiceException(string.Format("no 'value' attribute found in {0}/protocol_description/{1}fields/field[{2}]", protocolPath, tag, id));

			var isIdentifier = false;

			if (xeField.HasAttribute("identifier") && xeField.GetAttribute("identifier") == "true") {
				isIdentifier = true;
			}

			var field = new Field(id, xeField.GetAttribute("value"), isIdentifier);

			if (xeField.HasAttribute("type"))
				field.Type = xeField.GetAttribute("type");
			return field;
		}

		private static SQLStatement GetSQL(string protocolPath, ProtocolDescription desc, XmlElement xeDataSource, string tagName, string tagPath) {
			var tags = xeDataSource.GetElementsByTagName(tagName);

			if (tags == null || tags.Count == 0)
				throw new ServiceException(string.Format("no '{0}' element found in {1}/protocol_description/data_source", tagName, protocolPath));

			if (tags.Count > 1)
				throw new ServiceException(string.Format("too many '{0}' elements found in {1}/protocol_description/data_source", tagName, protocolPath));

			var xe = (XmlElement)tags[0];

			if (!xe.HasAttribute("name"))
				throw new ServiceException(string.Format("no 'name' attribute found in {0}/protocol_description/{1}/{2}", protocolPath, tagPath, tagName));

			if (!xe.HasAttribute("template"))
				throw new ServiceException(string.Format("no 'template' attribute found in {0}/protocol_description/{1}/{2}", protocolPath, tagPath, tagName));

			var sp = new SQLStatement(xe.GetAttribute("name"), xe.GetAttribute("template"));

			try {
				using (var sr = new StreamReader(string.Format("{0}\\{1}", PropertiesPath, sp.Template))) {
					_sql[sp.Name] = sr.ReadToEnd();
				}
			} catch {
				throw new ServiceException(string.Format("problem reading template {0}", sp.Template));
			}

			if (xe.HasChildNodes) {
				var parameters = GetParameters(protocolPath, desc.Data, string.Format("{0}/{1}", tagPath, tagName), xe);

				for (var i = 0; i < parameters.Count; ++i) {
					if (!parameters.ContainsKey(i))
						throw new ServiceException(string.Format("parameter with index {0} is missing in {1}/protocol_description/{2}/{3}", i, protocolPath, tagPath, tagName));

					sp.Parameters.Add(parameters[i]);
				}
			}
			return sp;
		}

		private static Collection GetCollection(string protocolPath, ProtocolDescription desc, XmlElement xeCollection) {
			var collId = xeCollection.GetAttribute("id");

			string delimiter = null;

			if (xeCollection.HasAttribute("delimiter"))
				delimiter = xeCollection.GetAttribute("delimiter");

			var dataSources = xeCollection.GetElementsByTagName("coll_data_source");

			if (dataSources == null || dataSources.Count == 0)
				throw new ServiceException(string.Format("no 'coll_data_source' element found in {0}/protocol_description/collections/collection[{1}]", protocolPath, collId));

			if (dataSources.Count > 1)
				throw new ServiceException(string.Format("too many 'coll_data_source' elements found in {0}/protocol_description/collections/collection[{1}]", protocolPath, collId));

			var xeDataSource = (XmlElement)dataSources[0];

			var sqlList = GetSQL(protocolPath, desc, xeDataSource, "list", string.Format("collections/collection[{0}]/coll_data_source", collId));

			var dsCollection = new DataSource(null, sqlList, null, null, null);

			if (!xeCollection.HasAttribute("multifetch"))
				throw new ServiceException(string.Format("no 'multifetch' attribute found in {{0}/protocol_description/collections/collection[{1}]", protocolPath, collId));

			var multifetch = bool.Parse(xeCollection.GetAttribute("multifetch"));

			var zeroPadRowNumber = 0;

			if (xeCollection.HasAttribute("rownumberzeropadding"))
				zeroPadRowNumber = int.Parse(xeCollection.GetAttribute("rownumberzeropadding"));

			var maxRowNumber = 0;

			if (xeCollection.HasAttribute("maxrownumber"))
				maxRowNumber = int.Parse(xeCollection.GetAttribute("maxrownumber"));

			var collection = new Collection(collId, dsCollection, delimiter, multifetch, zeroPadRowNumber, maxRowNumber);

			var xeCollectionFields = (XmlElement)xeCollection.GetElementsByTagName("coll_fields")[0];

			foreach (XmlElement xeCollectionField in xeCollectionFields.ChildNodes) {
				var field = GetField(protocolPath, xeCollectionField, collection.Fields, collection.Data, string.Format("collections/collection[{0}]/coll_fields", collId));

				collection.Data[field.Id] = field;
				collection.Fields[field.Id] = field;
			}

			if (collection.ZeroPadRowNumber > 0) {
				collection.Data["ROW_NUMBER"] = new RowNumber();
			}

			var xeCollectionOutput = (XmlElement)xeCollection.GetElementsByTagName("coll_output")[0];
			collection.Output.Template = xeCollectionOutput.GetAttribute("template");

			foreach (XmlElement xeNamedDataNode in xeCollectionOutput.ChildNodes) {
				switch (xeNamedDataNode.Name) {
					case "coll_data_template": {
							var dt = GetDataTemplate(protocolPath, collection.Data, xeNamedDataNode, string.Format("collections/collection[{0}]/coll_output/coll_data_template", collId));
							collection.Output.DataTemplates[dt.Id] = dt;
							collection.Data[dt.Id] = dt;
							collection.OrderedDataKeys.Add(dt.Id);
						}
						break;
					case "coll_conditional_data": {
							var cd = GetConditionalData(protocolPath, collection.Data, xeNamedDataNode, string.Format("collections/collection[{0}]/coll_output/coll_conditional_data", collId));
							collection.Output.ConditionalData[cd.Id] = cd;
							collection.Data[cd.Id] = cd;
							collection.OrderedDataKeys.Add(cd.Id);
						}
						break;
					case "coll_data_translation": {
							var dt = GetDataTranslation(protocolPath, collection.Data, xeNamedDataNode, string.Format("collections/collection[{0}]/coll_output/coll_data_translation", collId));
							collection.Output.DataTranslations[dt.Id] = dt;
							collection.Data[dt.Id] = dt;
							collection.OrderedDataKeys.Add(dt.Id);
						}
						break;
					case "coll_max_value": {
							var mv = GetMaxValue(protocolPath, collection.Data, xeNamedDataNode, string.Format("collections/collection[{0}]/coll_output/coll_max_value", collId));

							collection.Data[mv.Id] = mv;
							collection.Output.MaxValues[mv.Id] = mv;
							collection.OrderedDataKeys.Add(mv.Id);
						}
						break;
					case "coll_min_value": {
							var mv = GetMinValue(protocolPath, collection.Data, xeNamedDataNode, string.Format("collections/collection[{0}]/coll_output/coll_min_value", collId));

							collection.Data[mv.Id] = mv;
							collection.Output.MinValues[mv.Id] = mv;
							collection.OrderedDataKeys.Add(mv.Id);
						}
						break;
				}
			}

			try {
				using (var sr = new StreamReader(string.Format("{0}\\{1}", PropertiesPath, collection.Output.Template))) {
					_templates[collection.Output.Template] = sr.ReadToEnd();
				}
			} catch {
				throw new ServiceException(string.Format("problem reading template {0}", collection.Output.Template));
			}

			var additionalOutputList = xeCollection.GetElementsByTagName("coll_additional_output");

			foreach (XmlElement xeAdditionalOutput in additionalOutputList) {
				string additionalDelimiter = null;

				if (xeAdditionalOutput.HasAttribute("delimiter"))
					additionalDelimiter = xeAdditionalOutput.GetAttribute("delimiter");

				var additionalOutput = new AdditionalOutput(xeAdditionalOutput.GetAttribute("id"),
															xeAdditionalOutput.GetAttribute("template"),
															additionalDelimiter);
				collection.AdditionalOutputs[additionalOutput.Id] = additionalOutput;

				foreach (var dataKey in collection.Data.Keys) {
					additionalOutput.Data[dataKey] = collection.Data[dataKey];
				}

				foreach (XmlElement xeNamedDataNode in xeAdditionalOutput.ChildNodes) {
					switch (xeNamedDataNode.Name) {
						case "coll_data_template": {
								var dt = GetDataTemplate(protocolPath, additionalOutput.Data, xeNamedDataNode, string.Format("collections/collection[{0}]/coll_output/coll_data_template", collId));
								additionalOutput.DataTemplates[dt.Id] = dt;
								additionalOutput.Data[dt.Id] = dt;
								additionalOutput.OrderedDataKeys.Add(dt.Id);
							}
							break;
						case "coll_conditional_data": {
								var cd = GetConditionalData(protocolPath, additionalOutput.Data, xeNamedDataNode, string.Format("collections/collection[{0}]/coll_output/coll_conditional_data", collId));
								additionalOutput.ConditionalData[cd.Id] = cd;
								additionalOutput.Data[cd.Id] = cd;
								additionalOutput.OrderedDataKeys.Add(cd.Id);
							}
							break;
						case "coll_data_translation": {
								var dt = GetDataTranslation(protocolPath, additionalOutput.Data, xeNamedDataNode, string.Format("collections/collection[{0}]/coll_output/coll_data_translation", collId));
								additionalOutput.DataTranslations[dt.Id] = dt;
								additionalOutput.Data[dt.Id] = dt;
								additionalOutput.OrderedDataKeys.Add(dt.Id);
							}
							break;
						case "coll_max_value": {
								var mv = GetMaxValue(protocolPath, additionalOutput.Data, xeNamedDataNode, string.Format("collections/collection[{0}]/coll_output/coll_max_value", collId));

								additionalOutput.Data[mv.Id] = mv;
								additionalOutput.MaxValues[mv.Id] = mv;
								additionalOutput.OrderedDataKeys.Add(mv.Id);
							}
							break;
						case "coll_min_value": {
								var mv = GetMinValue(protocolPath, additionalOutput.Data, xeNamedDataNode, string.Format("collections/collection[{0}]/coll_output/coll_min_value", collId));
								additionalOutput.Data[mv.Id] = mv;
								additionalOutput.MinValues[mv.Id] = mv;
								additionalOutput.OrderedDataKeys.Add(mv.Id);
							}
							break;
					}
				}

				try {
					using (var sr = new StreamReader(string.Format("{0}\\{1}", PropertiesPath, additionalOutput.Template))) {
						_templates[additionalOutput.Template] = sr.ReadToEnd();
					}
				} catch {
					throw new ServiceException(string.Format("problem reading template {0}", additionalOutput.Template));
				}
			}
			return collection;
		}

		private static DataTemplate GetDataTemplate(string protocolPath, IDictionary<string, NamedData> data, XmlElement xeDataTemplate, string tag) {
			if (!xeDataTemplate.HasAttribute("id"))
				throw new ServiceException(string.Format("no 'id' attribute found in {0}/protocol_description/{1}[?]", protocolPath, tag));

			var id = xeDataTemplate.GetAttribute("id");

			if (!xeDataTemplate.HasAttribute("value"))
				throw new ServiceException(string.Format("no 'value' attribute found in {0}/protocol_description/{1}[{2}]", protocolPath, tag, id));

			var urlEncode = false;

			if (xeDataTemplate.HasAttribute("urlEncode"))
				urlEncode = (xeDataTemplate.GetAttribute("urlEncode") == "true");

			var dt = new DataTemplate(id, xeDataTemplate.GetAttribute("value"), urlEncode);

			if (xeDataTemplate.HasChildNodes) {
				var parameters = GetParameters(protocolPath, data, tag + "[" + id + "]", xeDataTemplate);

				for (var i = 0; i < parameters.Count; ++i) {
					if (!parameters.ContainsKey(i))
						throw new ServiceException(string.Format("parameter with index {0} is missing in {1}/protocol_description/{2}[{3}]", i, protocolPath, tag, id));

					dt.Parameters.Add(parameters[i]);
				}
			}
			return dt;
		}

		private static ConditionalData GetConditionalData(string protocolPath, IDictionary<string, NamedData> data, XmlElement xeConditionalData, string tag) {
			if (!xeConditionalData.HasAttribute("id"))
				throw new ServiceException(string.Format("no 'id' attribute found in {0}/protocol_description/{1}[?]", protocolPath, tag));

			var id = xeConditionalData.GetAttribute("id");

			if (!xeConditionalData.HasAttribute("source"))
				throw new ServiceException(string.Format("no 'source' attribute found in {0}/protocol_description/{1}[{2}]", protocolPath, tag, id));

			var source = xeConditionalData.GetAttribute("source");

			if (!data.ContainsKey(source))
				throw new ServiceException(string.Format("source {0} not found in data", source));

			var cd = new ConditionalData(id, data[source]);

			var xeConditionalOutput = (XmlElement)xeConditionalData.ChildNodes[0];
			cd.Output.Template = xeConditionalOutput.GetAttribute("template");

			foreach (XmlElement xeDataTemplate in xeConditionalOutput.GetElementsByTagName("conditional_data_template")) {
				var dt = GetDataTemplate(protocolPath, data, xeDataTemplate, string.Format("{0}[{1}]/conditional_data_template", tag, id));
				cd.Output.DataTemplates[dt.Id] = dt;
			}

			using (var sr = new StreamReader(string.Format("{0}\\{1}", PropertiesPath, cd.Output.Template))) {
				_templates[cd.Output.Template] = sr.ReadToEnd();
			}
			return cd;
		}

		private static DataTranslation GetDataTranslation(string protocolPath, IDictionary<string, NamedData> data, XmlElement xeDataTranslation, string tag) {
			if (!xeDataTranslation.HasAttribute("id"))
				throw new ServiceException(string.Format("no 'id' attribute found in {0}/protocol_description/{1}[?]", protocolPath, tag));

			var id = xeDataTranslation.GetAttribute("id");

			if (!xeDataTranslation.HasAttribute("source"))
				throw new ServiceException(string.Format("no 'source' attribute found in {0}/protocol_description/{1}[{2}]", protocolPath, tag, id));

			var source = xeDataTranslation.GetAttribute("source");

			if (!data.ContainsKey(source))
				throw new ServiceException(string.Format("source {0} not found in data", source));

			var dt = new DataTranslation(id, data[source]);

			foreach (XmlElement xeTranslation in xeDataTranslation.ChildNodes) {
				if (!xeTranslation.HasAttribute("source"))
					throw new ServiceException(string.Format("no 'source' attribute found in {0}/protocol_description/{1}[{2}]/translation[?]", protocolPath, tag, id));

				var trSource = OAIUtil.XmlEncode(xeTranslation.GetAttribute("source"));

				if (!xeTranslation.HasAttribute("target"))
					throw new ServiceException(string.Format("no 'target' attribute found in {0}/protocol_description/{1}[{2}]/translation[{3}]", protocolPath, tag, id, trSource));

				dt.Translations[trSource] = data[xeTranslation.GetAttribute("target")];
			}
			return dt;
		}

		private static MinValue GetMinValue(string protocolPath, IDictionary<string, NamedData> data, XmlElement xeMinValueNode, string tag) {
			if (!xeMinValueNode.HasAttribute("id"))
				throw new ServiceException(string.Format("no 'id' attribute found in {0}/protocol_description/{1}[?]", protocolPath, tag));

			var id = xeMinValueNode.GetAttribute("id");

			if (!xeMinValueNode.HasAttribute("type"))
				throw new ServiceException(string.Format("no 'type' attribute found in {0}/protocol_description/{1}[?]", protocolPath, tag));

			var type = xeMinValueNode.GetAttribute("type");

			switch (type) {
				case "date":
				case "int":
				case "long":
				case "float":
				case "double":
					break;
				default:
					throw new ServiceException(string.Format("non-supported 'type' {0} found in {1}/protocol_description/{2}[?]", type, protocolPath, tag));
			}

			var sourcesNodes = xeMinValueNode.ChildNodes;

			if (sourcesNodes == null || sourcesNodes.Count == 0)
				throw new ServiceException(string.Format("no 'source' elements found in {0}", string.Format("no 'type' attribute found in {0}/protocol_description/{1}[?]", protocolPath, tag)));

			var sources = new List<NamedData>();

			foreach (XmlElement sourceNode in sourcesNodes) {
				sources.Add(data[sourceNode.InnerText]);
			}

			return new MinValue(id, sources, type);
		}

		private static MaxValue GetMaxValue(string protocolPath, IDictionary<string, NamedData> data, XmlElement xeMaxValueNode, string tag) {
			if (!xeMaxValueNode.HasAttribute("id"))
				throw new ServiceException(string.Format("no 'id' attribute found in {0}/protocol_description/{1}[?]", protocolPath, tag));

			var id = xeMaxValueNode.GetAttribute("id");

			if (!xeMaxValueNode.HasAttribute("type"))
				throw new ServiceException(string.Format("no 'type' attribute found in {0}/protocol_description/{1}[?]", protocolPath, tag));

			var type = xeMaxValueNode.GetAttribute("type");

			switch (type) {
				case "date":
				case "int":
				case "long":
				case "float":
				case "double":
					break;
				default:
					throw new ServiceException(string.Format("non-supported 'type' {0} found in {1}/protocol_description/{2}[?]", type, protocolPath, tag));
			}

			var sourcesNodes = xeMaxValueNode.ChildNodes;

			if (sourcesNodes == null || sourcesNodes.Count == 0)
				throw new ServiceException(string.Format("no 'source' elements found in {0}", string.Format("no 'type' attribute found in {0}/protocol_description/{1}[?]", protocolPath, tag)));

			var sources = new List<NamedData>();

			foreach (XmlElement sourceNode in sourcesNodes) {
				sources.Add(data[sourceNode.InnerText]);
			}

			return new MaxValue(id, sources, type);
		}

		private static Dictionary<int, Parameter> GetParameters(string protocolPath, IDictionary<string, NamedData> data, string tag, XmlNode xe) {
			var parameters = new Dictionary<int, Parameter>();

			foreach (XmlElement xeParameter in xe.ChildNodes) {
				if (!xeParameter.HasAttribute("index"))
					throw new ServiceException(
						string.Format("no 'index' attribute found in {0}/protocol_description/{1}/parameter[?]",
									  protocolPath, tag));

				var index = int.Parse(xeParameter.GetAttribute("index"));

				if (parameters.ContainsKey(index))
					throw new ServiceException(
						string.Format("parameter with index {0} defined twice in {1}/protocol_description/{2}", index,
									  protocolPath, tag));

				if (!xeParameter.HasAttribute("source"))
					throw new ServiceException(
						string.Format("no 'source' attribute found in {0}/protocol_description/data_source/{1}/parameter[{2}]",
									  protocolPath, tag, index));

				var source = xeParameter.GetAttribute("source");

				if (!data.ContainsKey(source))
					throw new ServiceException(string.Format("source {0} not found in data", source));
				parameters[index] = new Parameter(index, data[source]);
			}
			return parameters;
		}
	}
}
